from __future__ import annotations

from math import ceil
from typing import Iterable

import torch
from escnn.group import Representation

from symm_learning.models.diffusion.cond_unet1d import SinusoidalPosEmb
from symm_learning.nn import IrrepSubspaceNormPooling, eAffine, eConv1d, eConvTranspose1d, eRMSNorm
from symm_learning.representation_theory import direct_sum


class _eChannelRMSNorm(torch.nn.Module):
    """Apply eRMSNorm over channels for tensors shaped (B, C, L)."""

    def __init__(self, rep: Representation, eps: float = 1e-6):
        super().__init__()
        self.norm = eRMSNorm(rep, eps=eps, equiv_affine=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        y = x.permute(0, 2, 1)  # (B, L, C)
        y = self.norm(y)
        return y.permute(0, 2, 1)


class eConditionalResidualBlock1D(torch.nn.Module):
    r"""Channel-equivariant conditional residual block with FiLM modulation.

    This block applies two equivariant convolutional layers with a residual
    connection. The output of the first convolutional block is modulated by a
    conditioning tensor using an equivariant FiLM (Feature-wise Linear Modulation)
    layer. The FiLM layer applies an equivariant affine transformation, where the
    scale and bias parameters are predicted by an invariant neural network from the
    conditioning tensor.

    The scale transformation is applied in the spectral domain, with a separate
    learnable parameter for each irreducible representation (irrep). The bias
    is applied only to the invariant (trivial) subspaces of the representation.
    See details in :class:`~symm_learning.nn.eAffine`.

    Functional constraint:

    .. math::
        f(\rho_{\mathcal{X}}(g)\mathbf{x},\,\rho_{\mathcal{C}}(g)\mathbf{c})
        = \rho_{\mathcal{Y}}(g)\,f(\mathbf{x},\mathbf{c}),
        \quad \forall g\in\mathbb{G}.
    """

    def __init__(
        self,
        in_rep: Representation,
        out_rep: Representation,
        cond_rep: Representation,
        kernel_size: int = 3,
        cond_predict_scale: bool = True,
        activation: torch.nn.Module = torch.nn.ReLU(),
        normalize: bool = True,
        init_scheme: str | None = "xavier_uniform",
    ):
        r"""Initialize the conditional residual block.

        Args:
            in_rep (:class:`~escnn.group.Representation`): Input representation :math:`\rho_{\text{in}}` acting on the
                channel axis.
            out_rep (:class:`~escnn.group.Representation`): Output representation :math:`\rho_{\text{out}}` for the
                convolutions and residual.
            cond_rep (:class:`~escnn.group.Representation`): Representation :math:`\rho_{\text{cond}}` of the
                conditioning vector used to predict FiLM parameters.
            kernel_size (int, optional): Spatial kernel size for the equivariant convolutions. Defaults to 3.
            cond_predict_scale (bool, optional): Whether the FiLM encoder predicts scale parameters.
                Currently must be ``True``. Defaults to True.
            activation (torch.nn.Module, optional): Non-linearity applied after each normalization. Defaults to ``ReLU``
            normalize (bool, optional): If ``True``, apply channel-wise equivariant RMS normalization. Defaults to True.
            init_scheme (str | None, optional): Weight initialization scheme for convolutions.
                Defaults to ``\"xavier_uniform\"``.
        """
        super().__init__()
        self.in_rep, self.out_rep, self.cond_rep = in_rep, out_rep, cond_rep
        if not cond_predict_scale:
            raise NotImplementedError("Currently only cond_predict_scale=True is supported.")

        self.conv1 = eConv1d(in_rep, out_rep, kernel_size, padding=kernel_size // 2, init_scheme=init_scheme)
        self.conv2 = eConv1d(out_rep, out_rep, kernel_size, padding=kernel_size // 2, init_scheme=init_scheme)

        self.norm1 = _eChannelRMSNorm(out_rep) if normalize else torch.nn.Identity()
        self.norm2 = _eChannelRMSNorm(out_rep) if normalize else torch.nn.Identity()
        self.act = activation

        self.affine = eAffine(in_rep=out_rep, learnable=False)
        self.film_dims = self.affine.num_scale_dof + self.affine.num_bias_dof

        # Invariant encoder of FiLM modulation parameters
        self.cond_encoder = torch.nn.Sequential(
            IrrepSubspaceNormPooling(in_rep=cond_rep),
            torch.nn.Linear(in_features=len(cond_rep.irreps), out_features=self.film_dims),
            torch.nn.Mish(),
        )

        self.residual_conv = (  # Final conv for residual addition if needed.
            eConv1d(in_rep, out_rep, 1, init_scheme=init_scheme) if in_rep != out_rep else torch.nn.Identity()
        )

    def forward(self, x: torch.Tensor, cond: torch.Tensor) -> torch.Tensor:
        """Apply two equivariant conv blocks with FiLM modulation and a residual connection.

        Args:
            x (torch.Tensor): Input of shape ``(B, in_rep.size, L)`` where ``B`` is batch and ``L`` is sequence length.
            cond (torch.Tensor): Conditioning tensor of shape ``(B, cond_rep.size)`` used to predict FiLM parameters.

        Returns:
            torch.Tensor: Output tensor of shape ``(B, out_rep.size, L)`` after modulation and residual addition.
        """
        assert x.shape[1] == self.in_rep.size, f"Expected channel dim {self.in_rep.size}, got {x.shape}"
        assert cond.shape[-1] == self.cond_rep.size, f"Cond dim mismatch {cond.shape}"

        B, _, L = x.shape

        out = self.conv1(x)  # (B, C, L)
        out = self.act(self.norm1(out))

        if self.film_dims > 0:
            film_params = self.cond_encoder(cond)  # (B, film_dims)
            scale_dof = torch.broadcast_to(
                film_params[:, None, : self.affine.num_scale_dof], (B, L, self.affine.num_scale_dof)
            )  # (B, num_scale_dof) -> (B, L, num_scale_dof)
            bias_dof = torch.broadcast_to(
                film_params[:, None, self.affine.num_scale_dof :], (B, L, self.affine.num_bias_dof)
            )  # (B, num_bias_dof) -> (B, L, num_bias_dof)
            out = self.affine(
                out.permute(0, 2, 1),  # (B, C, L) -> (B, L, C)
                scale_dof=scale_dof,
                bias_dof=bias_dof,
            ).permute(0, 2, 1)  # (B, L, C) -> (B, C, L)

        out = self.conv2(out)
        out = self.act(self.norm2(out))

        res = self.residual_conv(x) if self.residual_conv is not None else x
        return out + res

    @torch.no_grad()
    def check_equivariance(self, atol=1e-5, rtol=1e-5):  # noqa: D102
        G = self.in_rep.group
        B, L = 10, 30
        device, dtype = next(self.parameters()).device, next(self.parameters()).dtype
        x = torch.randn(B, self.in_rep.size, L, device=device, dtype=dtype)
        z = torch.randn(B, self.cond_rep.size, device=device, dtype=dtype)
        y = self(x, z)

        for _ in range(10):
            g = G.sample()
            rho_in = torch.tensor(self.in_rep(g), dtype=x.dtype, device=x.device)
            rho_out = torch.tensor(self.out_rep(g), dtype=y.dtype, device=y.device)
            rho_cond = torch.tensor(self.cond_rep(g), dtype=z.dtype, device=z.device)
            gx = torch.einsum("ij,bjl->bil", rho_in, x)
            gz = torch.einsum("ij,bj->bi", rho_cond, z)
            y_expected = self(gx, gz)
            gy = torch.einsum("ij,bjl->bil", rho_out, y)
            assert torch.allclose(gy, y_expected, atol=atol, rtol=rtol), (
                f"Equivariance failed for group element {g} with max error {(gy - y_expected).abs().max().item():.3e}"
            )


class eConditionalUnet1D(torch.nn.Module):
    r"""Equivariant U-Net for 1D signals with global conditioning and FiLM.

    The model defines:

    .. math::
        \mathbf{f}_{\mathbf{\theta}}:
        \mathcal{X}^{L} \times \mathcal{Z}_{\mathrm{local}}^{L} \times \mathcal{Z}_{\mathrm{global}} \times \mathbb{R}
        \to \mathcal{X}^{L}.

    Functional equivariance constraint:

    .. math::
        \mathbf{f}_{\mathbf{\theta}}(\rho_{\mathcal{X}}(g)\mathbf{x},\,\rho_{\mathcal{Z}}(g)\mathbf{z},\,t)
        = \rho_{\mathcal{X}}(g)\,\mathbf{f}_{\mathbf{\theta}}(\mathbf{x},\mathbf{z},t),
        \quad \forall g\in\mathbb{G}.
    """

    def __init__(
        self,
        in_rep: Representation,
        local_cond_rep: Representation | None,
        global_cond_rep: Representation | None = None,
        diffusion_step_embed_dim: int = 256,
        down_dims: Iterable[int] = (256, 512, 1024),
        kernel_size: int = 3,
        cond_predict_scale: bool = True,
        activation: torch.nn.Module = torch.nn.ReLU(),
        normalize: bool = True,
        downsample: str = "stride",
        init_scheme: str | None = "xavier_uniform",
    ):
        super().__init__()
        assert downsample in {"stride", "pooling"}, "downsample must be 'stride' or 'pooling'"
        self.in_rep = self.out_rep = in_rep
        self.global_cond_rep = global_cond_rep
        self.local_cond_rep = local_cond_rep
        self.downsample = downsample

        G = in_rep.group
        reg_rep = G.regular_representation
        trivial_rep = G.trivial_representation

        down_dims = list(down_dims)
        # all_dims = [in_rep.size] + down_dims
        reps = [in_rep] + [direct_sum([reg_rep] * ceil(d / reg_rep.size)) for d in down_dims]

        diffusion_step_encoder = torch.nn.Sequential(
            SinusoidalPosEmb(diffusion_step_embed_dim),
            torch.nn.Linear(diffusion_step_embed_dim, diffusion_step_embed_dim * 4),
            torch.nn.Mish(),
            torch.nn.Linear(diffusion_step_embed_dim * 4, diffusion_step_embed_dim),
        )

        if global_cond_rep is not None:
            cond_rep = direct_sum([global_cond_rep, direct_sum([trivial_rep] * diffusion_step_embed_dim)])
        else:
            cond_rep = direct_sum([trivial_rep] * diffusion_step_embed_dim)
        self.cond_rep = cond_rep
        self.diffusion_step_encoder = diffusion_step_encoder

        block_kwargs = dict(
            kernel_size=kernel_size,
            cond_predict_scale=cond_predict_scale,
            activation=activation,
            normalize=normalize,
            init_scheme=init_scheme,
        )
        conv_kwargs = dict(bias=True, init_scheme=init_scheme)

        in_out_reps = list(zip(reps[:-1], reps[1:]))

        local_cond_encoder = None
        if local_cond_rep is not None:
            first_out_rep = in_out_reps[0][1]
            local_cond_encoder = torch.nn.ModuleList(
                [
                    eConditionalResidualBlock1D(
                        in_rep=local_cond_rep,
                        out_rep=first_out_rep,
                        cond_rep=cond_rep,
                        **block_kwargs,
                    ),
                    eConditionalResidualBlock1D(
                        in_rep=local_cond_rep,
                        out_rep=first_out_rep,
                        cond_rep=cond_rep,
                        **block_kwargs,
                    ),
                ]
            )
        self.local_cond_encoder = local_cond_encoder

        mid_rep = reps[-1]
        self.mid_modules = torch.nn.ModuleList(
            [
                eConditionalResidualBlock1D(
                    mid_rep,
                    mid_rep,
                    cond_rep=cond_rep,
                    **block_kwargs,
                ),
                eConditionalResidualBlock1D(
                    mid_rep,
                    mid_rep,
                    cond_rep=cond_rep,
                    **block_kwargs,
                ),
            ]
        )

        down_modules = torch.nn.ModuleList()
        for idx, (rep_in, rep_out) in enumerate(in_out_reps):
            is_last = idx == len(in_out_reps) - 1
            down_modules.append(
                torch.nn.ModuleList(
                    [
                        eConditionalResidualBlock1D(
                            rep_in,
                            rep_out,
                            cond_rep=cond_rep,
                            **block_kwargs,
                        ),
                        eConditionalResidualBlock1D(
                            rep_out,
                            rep_out,
                            cond_rep=cond_rep,
                            **block_kwargs,
                        ),
                        eConv1d(
                            rep_out,
                            rep_out,
                            kernel_size=3,
                            stride=2 if not is_last and downsample == "stride" else 1,
                            padding=1,
                            **conv_kwargs,
                        )
                        if (downsample == "stride" and not is_last)
                        else (
                            torch.nn.MaxPool1d(kernel_size=2, stride=2)
                            if downsample == "pooling" and not is_last
                            else torch.nn.Identity()
                        ),
                    ]
                )
            )
        self.down_modules = down_modules

        up_modules = torch.nn.ModuleList()
        current_rep = in_out_reps[-1][1]  # deepest representation
        up_pairs = list(reversed(in_out_reps[1:]))  # skip shallowest pair
        for idx, (target_rep, skip_rep) in enumerate((pair[0], pair[1]) for pair in up_pairs):
            is_last = idx == len(up_pairs) - 1
            upsample_in_rep = direct_sum([current_rep, skip_rep])  # concat current + skip
            out_rep = target_rep
            up_modules.append(
                torch.nn.ModuleList(
                    [
                        eConditionalResidualBlock1D(
                            upsample_in_rep,
                            out_rep,
                            cond_rep=cond_rep,
                            **block_kwargs,
                        ),
                        eConditionalResidualBlock1D(
                            out_rep,
                            out_rep,
                            cond_rep=cond_rep,
                            **block_kwargs,
                        ),
                        eConvTranspose1d(
                            out_rep,
                            out_rep,
                            kernel_size=3,
                            stride=2 if not is_last and downsample == "stride" else 1,
                            padding=1,
                            **conv_kwargs,
                        )
                        if (downsample == "stride" and not is_last)
                        else (
                            torch.nn.Upsample(scale_factor=2, mode="nearest")
                            if downsample == "pooling" and not is_last
                            else torch.nn.Identity()
                        ),
                    ]
                )
            )
            current_rep = out_rep
        self.up_modules = up_modules

        first_rep = in_out_reps[0][1]
        final_norm = _eChannelRMSNorm(first_rep) if normalize else torch.nn.Identity()
        self.final_conv = torch.nn.Sequential(
            eConv1d(first_rep, first_rep, kernel_size=kernel_size, padding=kernel_size // 2, **conv_kwargs),
            final_norm,
            activation,
            eConv1d(first_rep, in_rep, kernel_size=1, padding=0, **conv_kwargs),
        )

    def forward(
        self,
        sample: torch.Tensor,
        timestep: torch.Tensor | float | int,
        local_cond: torch.Tensor | None = None,
        film_cond: torch.Tensor | None = None,
    ) -> torch.Tensor:
        """Run a forward pass of the equivariant U-Net.

        Args:
            sample (torch.Tensor): Input signal shaped ``(B, in_rep.size, L)``.
            timestep (torch.Tensor | float | int): Diffusion step; scalar or batch, broadcast to ``B``.
            local_cond (torch.Tensor | None, optional): Local conditioning signal shaped
                ``(B, local_cond_rep.size, L)`` when provided.
            film_cond (torch.Tensor | None, optional): Global conditioning vector shaped
                ``(B, global_cond_rep.size)`` to drive FiLM.

        Returns:
            torch.Tensor: Output tensor shaped ``(B, in_rep.size, L)``.
        """
        assert sample.shape[1] == self.in_rep.size, f"Expected channels {self.in_rep.size}, got {sample.shape}"
        device = sample.device

        t = timestep
        if not torch.is_tensor(t):
            t = torch.tensor([timestep], dtype=torch.long, device=device)
        elif t.ndim == 0:
            t = t[None].to(device)
        t = t.expand(sample.shape[0])

        film_diff = self.diffusion_step_encoder(t)
        film_feature = torch.cat([film_cond, film_diff], dim=-1) if film_cond is not None else film_diff
        assert film_feature.shape[-1] == self.cond_rep.size

        h_local = []
        if self.local_cond_encoder is not None and local_cond is not None:
            resnet_a, resnet_b = self.local_cond_encoder
            x_loc = resnet_a(local_cond, film_feature)
            h_local.append(x_loc)
            x_loc = resnet_b(local_cond, film_feature)
            h_local.append(x_loc)

        x = sample
        skips = []
        for idx, (res1, res2, down) in enumerate(self.down_modules):
            x = res1(x, film_feature)
            if idx == 0 and h_local:
                x = x + h_local[0]
            x = res2(x, film_feature)
            skips.append(x)
            x = down(x)

        for mid in self.mid_modules:
            x = mid(x, film_feature)

        for idx, (res1, res2, up) in enumerate(self.up_modules):
            skip = skips.pop()
            x = torch.cat([x, skip], dim=1)
            x = res1(x, film_feature)
            if idx == len(self.up_modules) - 1 and h_local:
                x = x + h_local[1]
            x = res2(x, film_feature)
            x = up(x)

        x = self.final_conv(x)
        return x

    @torch.no_grad()
    def check_equivariance(self, batch_size=3, length=5, atol=1e-5, rtol=1e-5):  # noqa: D102
        """Check equivariance under channel actions of the underlying fiber group."""
        G = self.in_rep.group
        device = next(self.parameters()).device
        dtype = next(self.parameters()).dtype

        x = torch.randn(batch_size, self.in_rep.size, length, device=device, dtype=dtype)
        local_cond = (
            torch.randn(batch_size, self.local_cond_rep.size, length, device=device, dtype=dtype)
            if self.local_cond_rep is not None
            else None
        )
        global_cond = (
            torch.randn(batch_size, self.global_cond_rep.size, device=device, dtype=dtype)
            if self.global_cond_rep is not None
            else None
        )
        t = torch.tensor(0, dtype=torch.long, device=device)

        y = self(x, timestep=t, local_cond=local_cond, film_cond=global_cond)

        for _ in range(10):
            g = G.sample()
            rho_in = torch.tensor(self.in_rep(g), dtype=dtype, device=device)
            gx = torch.einsum("ij,bjl->bil", rho_in, x)

            g_local = None
            if local_cond is not None:
                rho_local = torch.tensor(self.local_cond_rep(g), dtype=dtype, device=device)
                g_local = torch.einsum("ij,bjl->bil", rho_local, local_cond)

            g_global = None
            if global_cond is not None:
                rho_global = torch.tensor(self.global_cond_rep(g), dtype=dtype, device=device)
                g_global = torch.einsum("ij,bj->bi", rho_global, global_cond)

            y_expected = self(gx, timestep=t, local_cond=g_local, film_cond=g_global)

            rho_out = torch.tensor(self.out_rep(g), dtype=y.dtype, device=y.device)
            gy = torch.einsum("ij,bjl->bil", rho_out, y)
            assert torch.allclose(gy, y_expected, atol=atol, rtol=rtol), (
                f"Equivariance failed for group element {g} with max error {(gy - y_expected).abs().max().item():.3e}"
            )


if __name__ == "__main__":
    # Example usage
    from escnn.group import CyclicGroup

    diffusion_step_embed_dim = 32

    G = CyclicGroup(2)
    mx, my = 1, 2
    in_rep = direct_sum([G.regular_representation] * mx)
    out_rep = direct_sum([G.regular_representation] * my)
    cond_rep = direct_sum([G.regular_representation] * 2)

    print("Testing eConditionalResidualBlock1D ------------------------------------------")
    res_block = eConditionalResidualBlock1D(
        in_rep=in_rep,
        out_rep=out_rep,
        cond_rep=cond_rep,
        kernel_size=3,
        cond_predict_scale=True,
    )
    res_block.check_equivariance(atol=1e-5, rtol=1e-5)

    print("\nTesting eConditionalUnet1D ----------------------------------------------------")
    model = eConditionalUnet1D(
        in_rep=in_rep,
        local_cond_rep=None,
        global_cond_rep=cond_rep,
        diffusion_step_embed_dim=16,
        down_dims=[64, 128, 256],
        kernel_size=3,
        cond_predict_scale=True,
    )

    model.check_equivariance(atol=1e-5, rtol=1e-5)



from __future__ import annotations

import math
from typing import Iterable

import torch
from escnn.group import Representation

from symm_learning.models.emlp import eMLP, iMLP
from symm_learning.nn import eConv1d, eRMSNorm
from symm_learning.representation_theory import direct_sum


class _eChannelRMSNorm(torch.nn.Module):
    """Apply eRMSNorm over the channel dimension for tensors shaped (B, C, L)."""

    def __init__(self, rep: Representation, eps: float = 1e-6):
        super().__init__()
        self.rep = rep
        self.norm = eRMSNorm(rep, eps=eps, equiv_affine=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Normalize over channels; treat time steps as leading dimensions.
        y = x.permute(0, 2, 1)  # (B, L, C)
        y = self.norm(y)
        return y.permute(0, 2, 1)  # (B, C, L)


class eTimeCNNEncoder(torch.nn.Module):
    r"""Equivariant 1D CNN encoder built from channel-equivariant blocks.

    Inputs are plain tensors of shape ``(N, in_rep.size, H)``. Each conv block halves the
    time horizon via stride-2 convolution; optional eRMSNorm and pointwise activation follow.
    The flattened feature map feeds either an equivariant head (:class:`eMLP`) or an invariant
    head (:class:`iMLP`) depending on whether ``out_rep`` contains only the trivial irrep.

    The encoder defines:

    .. math::
        \mathbf{f}_{\mathbf{\theta}}: \mathcal{X}^{H} \to \mathcal{Y},

    where :math:`H` is the input horizon, :math:`\mathcal{X}` is the channel feature
    space transforming by :math:`\rho_{\mathcal{X}}`, and :math:`\mathcal{Y}` transforms
    by :math:`\rho_{\mathcal{Y}}`.

    Functional constraint (equivariant head):

    .. math::
        \mathbf{f}_{\mathbf{\theta}}(\rho_{\mathcal{X}}(g)\mathbf{x})
        = \rho_{\mathcal{Y}}(g)\mathbf{f}_{\mathbf{\theta}}(\mathbf{x})
        \quad \forall g\in\mathbb{G}.

    If ``out_rep`` is trivial-only, the head is invariant:

    .. math::
        \mathbf{f}_{\mathbf{\theta}}(\rho_{\mathcal{X}}(g)\mathbf{x})
        = \mathbf{f}_{\mathbf{\theta}}(\mathbf{x})
        \quad \forall g\in\mathbb{G}.
    """

    def __init__(
        self,
        in_rep: Representation,
        out_rep: Representation,
        hidden_channels: list[int],
        time_horizon: int,
        activation: torch.nn.Module = torch.nn.ReLU(),
        batch_norm: bool = False,
        bias: bool = True,
        mlp_hidden: list[int] = (128,),
        downsample: str = "stride",
        append_last_frame: bool = False,
        init_scheme: str | None = "xavier_uniform",
    ) -> None:
        r"""Create an equivariant time-series CNN encoder.

        Args:
            in_rep (:class:`~escnn.group.Representation`): Input representation :math:`\rho_{\text{in}}` defining the
                group action on the input channels.
            out_rep (:class:`~escnn.group.Representation`): Output representation :math:`\rho_{\text{out}}`. If it
                contains only trivial irreps, an :class:`iMLP` head is used; otherwise an :class:`eMLP` head is used.
            hidden_channels: List of output channel counts for each convolution block.
            time_horizon: Length of the input time series (number of frames).
            activation: Non-linearity applied after every convolution block.
            batch_norm: Whether to include channel-wise RMS normalization.
            bias: Whether to include bias in convolutions and linear heads.
            mlp_hidden: Hidden layer widths for the final MLP head.
            downsample: Downsampling strategy, either ``'stride'`` (stride-2 conv) or ``'pooling'`` (max pool).
            append_last_frame: Whether to concatenate the last frame of the input to the encoding before the head.
            init_scheme: Initialization scheme for equivariant layers.
        """
        super().__init__()
        assert len(hidden_channels) > 0, "At least one conv block is required"
        assert downsample in {"stride", "pooling"}, "downsample must be 'stride' or 'pooling'"

        self.in_rep = in_rep
        self.out_rep = out_rep
        self.time_horizon = int(time_horizon)
        self.append_last_frame = append_last_frame
        self.downsample = downsample

        G = in_rep.group
        reg_rep = G.regular_representation

        layers: list[torch.nn.Module] = []
        cnn_in_rep = in_rep
        h = self.time_horizon

        for c_out in hidden_channels:
            multiplicity = max(1, math.ceil(c_out / reg_rep.size))
            cnn_out_rep = direct_sum([reg_rep] * multiplicity)

            if self.downsample == "stride":
                layers.append(
                    eConv1d(
                        cnn_in_rep, cnn_out_rep, kernel_size=3, stride=2, padding=1, bias=bias, init_scheme=init_scheme
                    )
                )
                if batch_norm:
                    layers.append(_eChannelRMSNorm(cnn_out_rep))
                layers.append(activation)
                h = (h + 1) // 2
            else:  # pooling
                layers.append(
                    eConv1d(
                        cnn_in_rep, cnn_out_rep, kernel_size=3, stride=1, padding=1, bias=bias, init_scheme=init_scheme
                    )
                )
                if batch_norm:
                    layers.append(_eChannelRMSNorm(cnn_out_rep))
                layers.append(activation)
                layers.append(torch.nn.MaxPool1d(kernel_size=2, stride=2))
                h = h // 2

            cnn_in_rep = cnn_out_rep

        self.feature_layers = torch.nn.Sequential(*layers)
        assert h > 0, f"Horizon {self.time_horizon} too short for {len(hidden_channels)} blocks"
        self.time_horizon_out = h

        # Head input representation: repeat conv out_rep for each remaining time step
        head_rep = direct_sum([cnn_in_rep] * self.time_horizon_out)
        if self.append_last_frame:
            head_rep = direct_sum([head_rep, in_rep])
        self.head_in_rep = head_rep

        # Choose head: invariant if out_rep is trivial-only, else equivariant
        trivial_id = G.trivial_representation.id
        invariant_head = set(out_rep.irreps) == {trivial_id}
        if invariant_head:
            self.head = iMLP(
                in_rep=head_rep,
                out_dim=out_rep.size,
                hidden_units=list(mlp_hidden),
                activation=activation,
                bias=bias,
                init_scheme=init_scheme,
            )
        else:
            self.head = eMLP(
                in_rep=head_rep,
                out_rep=out_rep,
                hidden_units=list(mlp_hidden),
                activation=activation,
                bias=bias,
                init_scheme=init_scheme,
            )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Encode input of shape ``(N, in_rep.size, H)`` into ``(N, out_rep.size)``."""
        assert x.shape[-2:] == (self.in_rep.size, self.time_horizon), (
            f"Expected input shape (..., {self.in_rep.size}, {self.time_horizon}), got {x.shape}"
        )
        feats = self.feature_layers(x)
        z = feats.permute(0, 2, 1).reshape(feats.size(0), -1)
        if self.append_last_frame:
            z = torch.cat([z, x[:, :, -1]], dim=1)
        return self.head(z)

    @torch.no_grad()
    def check_equivariance(self, atol: float = 1e-5, rtol: float = 1e-5):
        """Check equivariance under channel actions of the underlying group."""
        import random

        G = self.in_rep.group
        B, L = 10, self.time_horizon
        dtype, device = next(self.head.parameters()).dtype, next(self.head.parameters()).device

        x = torch.randn(B, self.in_rep.size, L, device=device, dtype=dtype)
        y = self(x)

        elements = set(G.elements)
        for _ in range(10):
            g = random.choice(tuple(elements))
            rho_in = torch.tensor(self.in_rep(g), dtype=x.dtype, device=x.device)
            rho_out = torch.tensor(self.out_rep(g), dtype=y.dtype, device=y.device)
            gx = torch.einsum("ij,bjl->bil", rho_in, x)
            gy = self(gx)
            gy_exp = torch.einsum("ij,bj->bi", rho_out, y)
            assert torch.allclose(gy_exp, gy, atol=atol, rtol=rtol), (
                f"Equivariance failed for group element {g} with max error {(gy_exp - gy).abs().max().item():.3e}"
            )

            elements.remove(g)
            if len(elements) == 0:
                break



from __future__ import annotations

import torch


class _ChannelRMSNorm(torch.nn.Module):
    """Apply RMSNorm over channel dimension for inputs shaped (B, C, L)."""

    def __init__(self, channels: int, eps: float = 1e-6):
        super().__init__()
        self.norm = torch.nn.RMSNorm(channels, eps=eps)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # RMSNorm expects channels on the last dim.
        return self.norm(x.transpose(1, 2)).transpose(1, 2)


class TimeCNNEncoder(torch.nn.Module):
    r"""1D CNN baseline encoder for inputs of shape ``(N, in_dim, H)``.

    The model applies a stack of 1D conv blocks over time. Each block halves the time
    horizon using either stride-2 convolution or max-pooling. The final feature map of
    shape (N, C_L, H_out) is flattened and passed to a 1-hidden-layer MLP head.

    The encoder defines:

    .. math::
      \mathbf{f}_{\mathbf{\theta}}: \mathbb{R}^{d_{\mathrm{in}}\times H} \to \mathbb{R}^{d_{\mathrm{out}}}.

    This class does not impose group-equivariance constraints on channel transforms.
    It is an unconstrained baseline.

    Args:
      in_dim: Input channel dimension.
      out_dim: Output feature dimension.
      hidden_channels: Channels per conv block; depth equals len(hidden_channels).
      horizon: Input sequence length H.
      activation: Activation module or list (one per block). If a single module is given,
        it is replicated for all blocks.
      batch_norm: If True, add RMSNorm after each convolution.
      bias: Use bias in conv/linear layers.
      mlp_hidden: Hidden units of the final MLP head (list for deeper heads).
      downsample: "stride" (default) or "pooling"; each block halves H.
      append_last_frame: If True, concatenate x[:, :, -1] (N, in_dim) to flattened conv
        features before the MLP.

    Returns:
      Tensor of shape (N, out_dim).
    """

    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        hidden_channels: list[int],
        horizon: int,
        activation: torch.nn.Module = torch.nn.ReLU(),
        normalize: bool = False,
        bias: bool = True,
        mlp_hidden: list[int] = [128],
        downsample: str = "stride",
        append_last_frame: bool = False,
        init_scheme: str | None = "xavier_uniform",
    ) -> None:
        super().__init__()
        assert hasattr(hidden_channels, "__iter__") and hasattr(hidden_channels, "__len__"), (
            "hidden_channels must be a list of integers"
        )
        assert len(hidden_channels) > 0, "At least one conv block is required"
        assert downsample in {"stride", "pooling"}, "downsample must be 'stride' or 'pooling'"

        self.in_dim, self.out_dim = in_dim, out_dim
        self.horizon_in = int(horizon)
        self.batch_norm = normalize
        self.bias = bias
        self.downsample = downsample
        self.append_last_frame = append_last_frame

        conv_act = activation
        head_act = conv_act

        # Build conv feature extractor; kernel=3, padding=1 (time length preserved before downsampling)
        conv_layers = []
        cin = in_dim
        h = self.horizon_in

        for cout in hidden_channels:
            if self.downsample == "stride":
                # Conv1d with stride=2 halves time: L_out = floor((L+1)/2) for k=3,p=1
                conv = torch.nn.Conv1d(cin, cout, kernel_size=3, stride=2, padding=1, bias=bias)
                conv_layers.append(conv)
                if init_scheme is not None:
                    conv.reset_parameters()
                if normalize:
                    conv_layers.append(_ChannelRMSNorm(cout))
                conv_layers.append(conv_act)
                h = (h + 1) // 2
            else:  # pooling
                # Conv stride=1 keeps time, then MaxPool1d(2) halves: floor(L/2)
                conv = torch.nn.Conv1d(cin, cout, kernel_size=3, stride=1, padding=1, bias=bias)
                conv_layers.append(conv)
                if init_scheme is not None:
                    conv.reset_parameters()
                if normalize:
                    conv_layers.append(_ChannelRMSNorm(cout))
                conv_layers.append(conv_act)
                conv_layers.append(torch.nn.MaxPool1d(kernel_size=2, stride=2))
                h = h // 2

            cin = cout

        self.feature_extractor = torch.nn.Sequential(*conv_layers)
        self.horizon_out = max(1, h)
        flat_dim = cin * self.horizon_out
        if self.append_last_frame:
            flat_dim += in_dim

        # MLP head (supports multiple hidden layers)
        head_layers = []
        prev = flat_dim
        for hidden in mlp_hidden:
            head_layers.append(torch.nn.Linear(prev, hidden, bias=bias))
            # replicate a fresh activation instance of the same type as head_act
            head_layers.append(type(head_act)())
            prev = hidden
        head_layers.append(torch.nn.Linear(prev, out_dim, bias=bias))

        self.head = torch.nn.Sequential(*head_layers)
        torch.nn.init.orthogonal_(self.head[-1].weight)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass.

        :param x: Input tensor of shape ``(N, in_dim, H)``.
        :type x: torch.Tensor
        :returns: Encoded tensor of shape ``(N, out_dim)``.
        :rtype: torch.Tensor
        """
        z = self.feature_extractor(x)
        z = z.view(z.size(0), -1)
        if self.append_last_frame:
            last = x[:, :, -1]
            z = torch.cat([z, last], dim=1)
        return self.head(z)



"""Symmetric Learning - Machine Learning with Symmetry Priors.

**Symmetric Learning** is a machine learning library tailored to optimization problems
featuring symmetry priors. It provides equivariant neural network modules, models, and
utilities for leveraging group symmetries in data.

Modules
-------
nn
    Equivariant neural network layers (linear, conv, normalization, attention).
models
    Complete architectures (eMLP, eTransformer, eTimeCNNEncoder).
linalg
    Linear algebra utilities for symmetric vector spaces.
stats
    Statistics for symmetric random variables.
representation_theory
    Tools for group representations and irreducible decompositions.
"""

import logging

logger = logging.getLogger(__name__)

_deepcopy_patched = False


def _ensure_group_deepcopy_singleton() -> None:
    """Keep escnn Group/Representation singletons when modules are deep-copied.

    PyTorch clones prototype modules (e.g., in Transformer stacks) via ``copy.deepcopy``.
    escnn’s ``Group``/``Representation`` objects carry large caches (irreps, change-of-basis
    matrices) and are intended to be unique per group. Without overriding ``__deepcopy__``,
    every clone would duplicate those caches, ballooning memory and breaking the implicit
    singleton assumption. This helper installs a no-op ``__deepcopy__`` on both classes,
    ensuring all clones reuse the same underlying instances and shared caches.
    """
    global _deepcopy_patched
    if _deepcopy_patched:
        return

    from escnn.group import Group, GroupElement, HomSpace, IrreducibleRepresentation, Representation

    def _identity_deepcopy(self, memo):
        """No-op deepcopy to keep singleton instances."""
        logger.debug(f"Reusing singleton {self.__class__.__name__} during deepcopy")
        memo[id(self)] = self
        return self

    for _cls in (Group, Representation, IrreducibleRepresentation, GroupElement, HomSpace):
        if not hasattr(_cls, "__deepcopy__"):
            _cls.__deepcopy__ = _identity_deepcopy  # type: ignore[attr-defined]
    _deepcopy_patched = True


# Patch eagerly on package import so any module using escnn reps shares the singleton caches.
_ensure_group_deepcopy_singleton()

from symm_learning import linalg, models, nn, representation_theory, stats, utils  # noqa: F401

__all__ = [
    "linalg",
    "models",
    "nn",
    "representation_theory",
    "stats",
    "utils",
]



"""Symmetric Learning - Linear Algebra Utilities.

Utility functions for linear algebra operations on symmetric vector spaces with known
group representations.

Functions
---------
lstsq
    Least squares solution constrained to equivariant linear maps.
invariant_orthogonal_projector
    Orthogonal projection onto the G-invariant subspace.
irrep_radii
    Compute Euclidean radius of each irreducible subspace.
isotypic_signal2irreducible_subspaces
    Flatten isotypic signals into irreducible subspace components.
"""

import numpy as np
import torch
from escnn.group import Representation
from torch import Tensor


def isotypic_signal2irreducible_subspaces(x: Tensor, rep_x: Representation):
    r"""Flatten an isotypic signal into its irreducible-subspace coordinates.

    This function assumes :math:`\mathcal{X}` is a single isotypic subspace of type :math:`k`, i.e.

    .. math::
        \rho_{\mathcal{X}} = \bigoplus_{i\in[1,n_k]} \hat{\rho}_k.

    For an input :math:`\mathbf{x}` of shape :math:`(n, n_k \cdot d_k)`, where :math:`d_k=\dim(\hat{\rho}_k)`, the
    output rearranges coordinates to shape :math:`(n \cdot d_k, n_k)` so each column stores one irrep copy across the
    sample axis.

    .. math::
        \mathbf{z}[:, i] = [x_{1,i,1}, \ldots, x_{1,i,d_k}, x_{2,i,1}, \ldots, x_{n,i,d_k}]^\top.

    Args:
        x (:class:`torch.Tensor`): Shape :math:`(n, n_k \cdot d_k)`.
        rep_x (:class:`~escnn.group.Representation`): Representation in isotypic basis with a single active irrep type.

    Returns:
        Tensor: Flattened irreducible-subspace signal of shape :math:`(n \cdot d_k, n_k)`.

    Shape:
        :math:`(n \cdot d_k, n_k)`.
    """
    assert len(rep_x._irreps_multiplicities) == 1, "Random variable is assumed to be in a single isotypic subspace."
    irrep_id = rep_x.irreps[0]
    irrep_dim = rep_x.group.irrep(*irrep_id).size
    mk = rep_x._irreps_multiplicities[irrep_id]  # Multiplicity of the irrep in X

    Z = x.view(-1, mk, irrep_dim).permute(0, 2, 1).reshape(-1, mk)

    assert Z.shape == (x.shape[0] * irrep_dim, mk)
    return Z


def irrep_radii(x: Tensor, rep: Representation) -> Tensor:
    r"""Compute Euclidean radii for all irreducible-subspace features.

    Let :math:`\rho_{\mathcal{X}}` be the (possibly decomposable) representation of a vector space
    :math:`\mathcal{X}`:

    .. math::
        \rho_{\mathcal{X}} = \mathbf{Q}\left(
        \bigoplus_{k\in[1,n_{\text{iso}}]}
        \bigoplus_{i\in[1,n_k]}
        \hat{\rho}_k
        \right)\mathbf{Q}^T.

    We first change to the irrep-spectral basis induced by this
    :ref:`isotypic decomposition <isotypic-decomposition-example>`
    (as returned by :func:`~symm_learning.representation_theory.isotypic_decomp_rep`),
    :math:`\hat{\mathbf{x}}=\mathbf{Q}^T\mathbf{x}`, and then compute the radius of each irrep copy:

    .. math::
        r_{k,i} = \lVert \hat{\mathbf{x}}_{k,i} \rVert_2.

    Args:
        x: (:class:`torch.Tensor`) of shape :math:`(..., D)` describing vectors transforming according to ``rep``.
        rep: (:class:`~escnn.group.Representation`) acting on the last dimension of ``x``.

    Returns:
        (:class:`torch.Tensor`): Radii of shape :math:`(..., N)` where :math:`N=\texttt{len(rep.irreps)}`. The output
        order follows :attr:`rep.irreps` (one radius per irreducible copy in the decomposition).

    Shape:
        - **Input** ``x``: :math:`(..., D)` with :math:`D=\dim(\rho_{\mathcal{X}})`.
        - **Output**: :math:`(..., N)` containing the per-irrep Euclidean norms.

    Note:
        For repeated calls with the same representation object ``rep``, the matrix
        :math:`\mathbf{Q}^{-1}` is cached in ``rep.attributes["Q_inv"]`` and reused.
    """
    if x.shape[-1] != rep.size:
        raise ValueError(f"Expected last dimension {rep.size}, got {x.shape[-1]}")

    if "Q_inv" not in rep.attributes:  # cache inverse change-of-basis for reuse
        Q_inv = torch.tensor(rep.change_of_basis_inv, device=x.device, dtype=x.dtype)
        rep.attributes["Q_inv"] = Q_inv
    else:
        Q_inv = rep.attributes["Q_inv"].to(x.device, x.dtype)

    # Change to irrep-spectral basis
    x_spectral = torch.einsum("ij,...j->...i", Q_inv, x)
    # Compute a mask for each irreducible subspace (subspace associated to an irrep)
    n_subspaces = len(rep.irreps)
    subspace_ids = _get_irrep_subspace_index(rep).to(x.device)

    flat = x_spectral.reshape(-1, rep.size)
    flat_sq = flat.pow(2)  # squared magnitudes per coordinate

    scatter_idx = subspace_ids.view(1, -1).expand(flat_sq.size(0), -1)  # broadcast labels across batch
    radii_sq = torch.zeros(flat_sq.size(0), n_subspaces, device=x.device, dtype=x.dtype)
    radii_sq.scatter_add_(1, scatter_idx, flat_sq)  # sum squares inside each irrep block

    radii = radii_sq.sqrt().reshape(*x_spectral.shape[:-1], n_subspaces)
    return radii


def lstsq(x: Tensor, y: Tensor, rep_x: Representation, rep_y: Representation):
    r"""Computes a solution to the least squares problem of a system of linear equations with equivariance constraints.

    The :math:`\mathbb{G}`-equivariant least squares problem to the linear system of equations
    :math:`\mathbf{Y} = \mathbf{A}\,\mathbf{X}`, is defined as:

    .. math::
        \begin{align}
            &\operatorname{argmin}_{\mathbf{A}} \| \mathbf{Y} - \mathbf{A}\,\mathbf{X} \|_F \\
            & \text{s.t.} \quad \rho_{\mathcal{Y}}(g) \mathbf{A} = \mathbf{A}\rho_{\mathcal{X}}(g) \quad \forall g
            \in \mathbb{G},
        \end{align}

    where :math:`\mathbf{X}: \Omega \to \mathcal{X}` and :math:`\mathbf{Y}: \Omega \to \mathcal{Y}` are random
    variables taking values in representation spaces :math:`\mathcal{X}` and :math:`\mathcal{Y}`, and
    :math:`\rho_{\mathcal{X}}`, :math:`\rho_{\mathcal{Y}}` are the corresponding (possibly decomposable)
    representations of :math:`\mathbb{G}`.

    Args:
        x (:class:`torch.Tensor`): Realizations of the random variable :math:`\mathbf{X}` with shape
            :math:`(N, D_x)`, where :math:`N` is the number of samples.
        y (:class:`torch.Tensor`):
            Realizations of the random variable :math:`\mathbf{Y}` with shape :math:`(N, D_y)`.
        rep_x (:class:`~escnn.group.Representation`):
            Representation :math:`\rho_{\mathcal{X}}` acting on the vector space :math:`\mathcal{X}`.
        rep_y (:class:`~escnn.group.Representation`):
            Representation :math:`\rho_{\mathcal{Y}}` acting on the vector space :math:`\mathcal{Y}`.

    Returns:
        :class:`torch.Tensor`:
            A :math:`(D_y \times D_x)` matrix :math:`\mathbf{A}` satisfying the :math:`\mathbb{G}`-equivariance
            constraint and minimizing :math:`\|\mathbf{Y} - \mathbf{A}\,\mathbf{X}\|^2`.

    Shape:
        - X: :math:`(N, D_x)`
        - Y: :math:`(N, D_y)`
        - Output: :math:`(D_y, D_x)`

    Note:
        This function calls :func:`~symm_learning.representation_theory.isotypic_decomp_rep`, which caches
        decompositions in the group representation registry. Repeated calls with the same representations reuse
        cached decompositions.
    """
    from symm_learning.representation_theory import isotypic_decomp_rep

    #   assert X.shape[0] == Y.shape[0], "Expected equal number of samples in X and Y"
    assert x.shape[1] == rep_x.size, f"Expected X shape (N, {rep_x.size}), got {x.shape}"
    assert y.shape[1] == rep_y.size, f"Expected Y shape (N, {rep_y.size}), got {y.shape}"
    assert x.shape[-1] == rep_x.size, f"Expected X shape (..., {rep_x.size}), got {x.shape}"
    assert y.shape[-1] == rep_y.size, f"Expected Y shape (..., {rep_y.size}), got {y.shape}"

    rep_X_iso = isotypic_decomp_rep(rep_x)
    rep_Y_iso = isotypic_decomp_rep(rep_y)
    # Changes of basis from the Disentangled/Isotypic-basis of X, and Y to the original basis.
    Qx = torch.tensor(rep_X_iso.change_of_basis, device=x.device, dtype=x.dtype)
    Qy = torch.tensor(rep_Y_iso.change_of_basis, device=y.device, dtype=y.dtype)
    rep_X_iso_subspaces = rep_X_iso.attributes["isotypic_reps"]
    rep_Y_iso_subspaces = rep_Y_iso.attributes["isotypic_reps"]

    # Get the dimensions of the isotypic subspaces of the same type in the input/output representations.
    iso_idx_X, iso_idx_Y = {}, {}
    x_dim = 0
    for iso_id, rep_k in rep_X_iso_subspaces.items():
        iso_idx_X[iso_id] = slice(x_dim, x_dim + rep_k.size)
        x_dim += rep_k.size
    y_dim = 0
    for iso_id, rep_k in rep_Y_iso_subspaces.items():
        iso_idx_Y[iso_id] = slice(y_dim, y_dim + rep_k.size)
        y_dim += rep_k.size

    x_iso = torch.einsum("ij,...j->...i", Qx.T, x)
    y_iso = torch.einsum("ij,...j->...i", Qy.T, y)
    A_iso = torch.zeros((rep_y.size, rep_x.size), dtype=x.dtype, device=x.device)
    for iso_id in rep_Y_iso_subspaces:
        if iso_id not in rep_X_iso_subspaces:
            continue  # No covariance between the isotypic subspaces of different types.
        x_k = x_iso[..., iso_idx_X[iso_id]]
        y_k = y_iso[..., iso_idx_Y[iso_id]]
        rep_X_k = rep_X_iso_subspaces[iso_id]
        rep_Y_k = rep_Y_iso_subspaces[iso_id]
        # Compute empirical least-squares.
        A_k_emp = torch.linalg.lstsq(x_k, y_k).solution.T
        A_k = _project_to_irrep_endomorphism_basis(A_k_emp, rep_X_k, rep_Y_k)
        A_iso[iso_idx_Y[iso_id], iso_idx_X[iso_id]] = A_k

    # Change back to the original input output basis sets
    A = Qy @ A_iso @ Qx.T
    return A


def invariant_orthogonal_projector(rep_x: Representation) -> Tensor:
    r"""Computes the orthogonal projection to the invariant subspace.

    The input representation :math:`\rho_{\mathcal{X}}: \mathbb{G} \mapsto \mathbb{G}\mathbb{L}(\mathcal{X})` is
    transformed to the spectral basis given by:

    .. math::
        \rho_{\mathcal{X}} = \mathbf{Q}\left(
        \bigoplus_{k\in[1,n_{\text{iso}}]}
        \bigoplus_{i\in[1,n_k]}
        \hat{\rho}_k
        \right)\mathbf{Q}^T

    where :math:`\hat{\rho}_k` are irreducible representations of :math:`\mathbb{G}`, :math:`n_k` is the multiplicity
    of type :math:`k`, and :math:`\mathbf{Q}: \mathcal{X}\to\mathcal{X}` is the orthogonal change of basis from the
    irrep-spectral basis to the original basis.

    Define the diagonal selector :math:`\mathbf{S}\in\mathbb{R}^{D\times D}` in irrep-spectral coordinates by

    .. math::
        S_{jj} = \begin{cases}
        1, & \text{if coordinate } j \text{ belongs to a trivial irrep copy}, \\
        0, & \text{otherwise}.
        \end{cases}

    Then the orthogonal projector onto the invariant subspace
    :math:`\mathcal{X}^{\text{inv}}=\{\mathbf{x}\in\mathcal{X}: \rho_{\mathcal{X}}(g)\mathbf{x}=\mathbf{x},
    \forall g\in\mathbb{G}\}` is

    .. math::
        \mathbf{P}_{\mathrm{inv}} = \mathbf{Q}\,\mathbf{S}\,\mathbf{Q}^T.

    This projector enforces the invariance constraint:

    .. math::
        \rho_{\mathcal{X}}(g)\,\mathbf{P}_{\mathrm{inv}}
        = \mathbf{P}_{\mathrm{inv}}\,\rho_{\mathcal{X}}(g)
        = \mathbf{P}_{\mathrm{inv}} \quad \forall g\in\mathbb{G}.

    Args:
        rep_x (:class:`~escnn.group.Representation`): The representation for which the orthogonal projection to the
        invariant subspace is computed.

    Returns:
        :class:`torch.Tensor`: The orthogonal projection matrix to the invariant subspace,
        :math:`\mathbf{Q} \mathbf{S} \mathbf{Q}^T`.
    """
    Qx_T, Qx = Tensor(rep_x.change_of_basis_inv), Tensor(rep_x.change_of_basis)

    # S is an indicator of which dimension (in the irrep-spectral basis) is associated with a trivial irrep
    S = torch.zeros((rep_x.size, rep_x.size))
    irreps_dimension = []
    cum_dim = 0
    for irrep_id in rep_x.irreps:
        irrep = rep_x.group.irrep(*irrep_id)
        # Get dimensions of the irrep in the original basis
        irrep_dims = range(cum_dim, cum_dim + irrep.size)
        irreps_dimension.append(irrep_dims)
        if irrep_id == rep_x.group.trivial_representation.id:
            # this dimension is associated with a trivial irrep
            S[irrep_dims, irrep_dims] = 1
        cum_dim += irrep.size

    inv_projector = Qx @ S @ Qx_T
    return inv_projector


def _project_to_irrep_endomorphism_basis(A: Tensor, rep_x: Representation, rep_y: Representation) -> Tensor:
    r"""Projects a linear map A: X -> Y between two isotypic spaces to the space of equivariant linear maps.

    Given a linear map :math:`A: X \to Y`, where :math:`X` and :math:`Y` are isotypic vector spaces of the same type,
    that is, their representations are built from :math:`m_x` and :math:`m_y` copies of the same irrep, this
    function projects the linear map to the space of equivariant linear maps between the two isotypic spaces.

    Args:
        A (:class:`torch.Tensor`): The linear map to be projected, of shape :math:`(m_y \cdot d, m_x \cdot d)`,
            where :math:`d` is the dimension of the irreducible representation, and :math:`m_x` and :math:`m_y`
            are the multiplicities of the irreducible representation in :math:`X` and :math:`Y`, respectively.
        rep_x (:class:`~escnn.group.Representation`): The representation of the isotypic space :math:`X`.
        rep_y (:class:`~escnn.group.Representation`): The representation of the isotypic space :math:`Y`.

    Returns:
        A_equiv (:class:`torch.Tensor`): A projected linear map of shape :math:`(m_y \cdot d, m_x \cdot d)` which
            commutes with the action of the group on the isotypic spaces :math:`X` and :math:`Y`. That is:
            :math:`A_{equiv} \circ \rho_X(g) = \rho_Y(g) \circ A_{equiv}` for all :math:`g \in \mathbb{G}`.
    """
    irrep_id = rep_x.irreps[0]
    irrep = rep_x.group.irrep(*irrep_id)
    assert A.shape == (rep_y.size, rep_x.size), "Expected A: X -> Y"
    assert len(rep_x._irreps_multiplicities) == 1, f"Expected rep with a single irrep type, got {rep_x.irreps}"
    assert len(rep_y._irreps_multiplicities) == 1, f"Expected rep with a single irrep type, got {rep_y.irreps}"
    assert irrep_id == rep_y.irreps[0], f"Irreps {irrep_id} != {rep_y.irreps[0]}. Hence A=0"
    x_in_iso_basis = np.allclose(rep_x.change_of_basis_inv, np.eye(rep_x.size), atol=1e-6, rtol=1e-4)
    assert x_in_iso_basis, "Expected X to be in spectral/isotypic basis"
    y_in_iso_basis = np.allclose(rep_y.change_of_basis_inv, np.eye(rep_y.size), atol=1e-6, rtol=1e-4)
    assert y_in_iso_basis, "Expected Y to be in spectral/isotypic basis"

    m_x = rep_x._irreps_multiplicities[irrep_id]  # Multiplicity of the irrep in X
    m_y = rep_y._irreps_multiplicities[irrep_id]  # Multiplicity of the irrep in Y

    # Get the basis of endomorphisms of the irrep (B, d, d)  B = 1 | 2 | 4
    irrep_end_basis = torch.tensor(irrep.endomorphism_basis(), device=A.device, dtype=A.dtype)
    A_irreps = A.view(m_y, irrep.size, m_x, irrep.size).permute(0, 2, 1, 3).contiguous()
    # Compute basis expansion coefficients of each irrep cross-covariance in basis of End(irrep) ========
    # Frobenius inner product  <C , Ψ_b>  =  Σ_{i,j} C_{ij} Ψ_b,ij
    A_irreps_basis_coeff = torch.einsum("mnij,bij->mnb", A_irreps, irrep_end_basis)  # (m_y , m_x , B)
    # squared norms ‖Ψ_b‖² (only once, very small)
    basis_coeff_norms = torch.einsum("bij,bij->b", irrep_end_basis, irrep_end_basis)  # (B,)
    A_irreps_basis_coeff = A_irreps_basis_coeff / basis_coeff_norms[None, None]

    A_irreps = torch.einsum("...b,bij->...ij", A_irreps_basis_coeff, irrep_end_basis)  # (m_y , m_x , d , d)
    # Reshape to (my * d, mx * d)
    A_equiv = A_irreps.permute(0, 2, 1, 3).reshape(m_y * irrep.size, m_x * irrep.size)
    return A_equiv


def _get_irrep_subspace_index(rep: Representation):
    labels = torch.empty(rep.size, dtype=torch.long)
    irreps_dims = {id: rep.group.irrep(*id).size for id in rep._irreps_multiplicities}
    pos = 0
    for count, irrep_id in enumerate(rep.irreps):
        labels[pos : pos + irreps_dims[irrep_id]] = count  # fill contiguous block for this copy
        pos += irreps_dims[irrep_id]

    return labels



from __future__ import annotations

import logging
from collections.abc import Callable
from math import ceil
from typing import Literal, Optional, Union

import torch
import torch.nn.functional as F
from escnn.group import Representation
from torch import Tensor

# from torch.nn import Transformer
from symm_learning.nn.activation import eMultiheadAttention
from symm_learning.nn.linear import eLinear
from symm_learning.nn.normalization import eLayerNorm, eRMSNorm
from symm_learning.representation_theory import direct_sum

logger = logging.getLogger(__name__)


class eTransformerEncoderLayer(torch.nn.Module):
    r"""Equivariant Transformer encoder layer with the same API as ``torch.nn.TransformerEncoderLayer``.

    Applies :class:`eMultiheadAttention` followed by an equivariant feed-forward block
    built from :class:`eLinear` layers and :class:`eLayerNorm`, mirroring PyTorch’s ordering
    (pre- or post-norm) while constraining every linear map to commute with the group action.

    The layer defines:

    .. math::
        \mathbf{f}_{\mathbf{\theta}}: \mathcal{X}^{T} \to \mathcal{X}^{T}.

    Functional equivariance constraint:

    .. math::
        \mathbf{f}_{\mathbf{\theta}}(\rho_{\mathcal{X}}(g)\mathbf{x}_{1:T})
        = \rho_{\mathcal{X}}(g)\,\mathbf{f}_{\mathbf{\theta}}(\mathbf{x}_{1:T})
        \quad \forall g\in\mathbb{G},

    where :math:`\rho_{\mathcal{X}}(g)` acts on the feature/channel axis at every token.
    """

    __constants__ = ["norm_first"]

    def __init__(
        self,
        in_rep: Representation,
        nhead: int,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: Union[str, Callable[[Tensor], Tensor]] = F.relu,
        layer_norm_eps: float = 1e-5,
        batch_first: bool = True,
        norm_first: bool = True,
        norm_module: Literal["layernorm", "rmsnorm"] = "rmsnorm",
        bias: bool = True,
        device=None,
        dtype=None,
        init_scheme: str | None = "xavier_uniform",
    ) -> None:
        r"""Create an equivariant Transformer encoder layer.

        Args:
            in_rep (:class:`~escnn.group.Representation`): Input representation :math:`\rho_{\text{in}}`.
            nhead: Number of attention heads.
            dim_feedforward: Hidden dimension of the feedforward network.
            dropout: Dropout probability.
            activation: Activation function (``'relu'`` or ``'gelu'``).
            layer_norm_eps: Epsilon for layer normalization.
            batch_first: If ``True``, input/output shape is ``(B, T, D)``.
            norm_first: If ``True``, apply normalization before attention/feedforward.
            norm_module: Normalization layer type (``'layernorm'`` or ``'rmsnorm'``).
            bias: Whether to use bias in linear layers.
            device: Tensor device.
            dtype: Tensor dtype.
            init_scheme: Initialization scheme for equivariant layers.
        """
        super().__init__()
        if dim_feedforward <= 0:
            raise ValueError(f"dim_feedforward must be positive, got {dim_feedforward}")

        self.in_rep, self.out_rep = in_rep, in_rep
        factory_kwargs = {"device": device, "dtype": dtype or torch.get_default_dtype()}

        G = in_rep.group
        num_hidden_reps = max(1, ceil(dim_feedforward / G.order()))
        self.embedding_rep = direct_sum([G.regular_representation] * num_hidden_reps)
        self.hidden_dim = self.embedding_rep.size
        self.requested_dim_feedforward = dim_feedforward

        self.self_attn = eMultiheadAttention(
            in_rep=self.in_rep,
            num_heads=nhead,
            dropout=dropout,
            bias=bias,
            batch_first=batch_first,
            device=device,
            dtype=dtype,
            init_scheme=init_scheme,
        )

        self.linear1 = eLinear(self.in_rep, self.embedding_rep, bias, init_scheme=init_scheme).to(**factory_kwargs)
        self.linear2 = eLinear(self.embedding_rep, self.out_rep, bias, init_scheme=init_scheme).to(**factory_kwargs)

        self.dropout = torch.nn.Dropout(dropout)
        self.dropout1 = torch.nn.Dropout(dropout)
        self.dropout2 = torch.nn.Dropout(dropout)

        if norm_module == "layernorm":
            norm_cls = eLayerNorm
            norm_kwargs = {"bias": bias} | factory_kwargs
            raise ValueError("eLayerNorm is numerically unstable. Use eRMSNorm instead for now.")
        elif norm_module == "rmsnorm":
            norm_cls = eRMSNorm
            norm_kwargs = factory_kwargs
        else:
            raise ValueError(f"norm_module must be 'layernorm' or 'rmsnorm', got {norm_module}")

        self.norm1 = norm_cls(self.in_rep, eps=layer_norm_eps, equiv_affine=True, **norm_kwargs)
        self.norm2 = norm_cls(self.out_rep, eps=layer_norm_eps, equiv_affine=True, **norm_kwargs)

        self.norm_first = norm_first
        self.batch_first = batch_first

        if isinstance(activation, str):
            activation = _get_activation_fn(activation)
        self.activation = activation

        if init_scheme is not None:
            self.reset_parameters(scheme=init_scheme)

    def forward(
        self,
        src: Tensor,
        src_mask: Tensor = None,
        src_key_padding_mask: Tensor = None,
        is_causal: bool = False,
    ) -> Tensor:
        r"""Pass the input through the equivariant encoder layer.

        Args:
            src: input sequence of shape ``(T, B, D)`` or ``(B, T, D)`` depending on ``batch_first``,
                with last dimension equal to ``in_rep.size``.
            src_mask: optional attention mask for the input sequence.
            src_key_padding_mask: optional padding mask for the batch.
            is_causal: if ``True``, applies a causal mask to the self-attention block.
        """
        src_key_padding_mask = F._canonical_mask(
            mask=src_key_padding_mask,
            mask_name="src_key_padding_mask",
            other_type=F._none_or_dtype(src_mask),
            other_name="src_mask",
            target_type=src.dtype,
        )
        src_mask = F._canonical_mask(
            mask=src_mask,
            mask_name="src_mask",
            other_type=None,
            other_name="",
            target_type=src.dtype,
            check_other=False,
        )

        x = src
        if self.norm_first:
            x = x + self._self_attention_block(self.norm1(x), src_mask, src_key_padding_mask, is_causal=is_causal)
            x = x + self._feed_forward_block(self.norm2(x))
        else:
            x = self.norm1(x + self._self_attention_block(x, src_mask, src_key_padding_mask, is_causal=is_causal))
            x = self.norm2(x + self._feed_forward_block(x))

        return x

    def _self_attention_block(
        self, x: Tensor, attn_mask: Tensor = None, key_padding_mask: Tensor = None, is_causal: bool = False
    ) -> Tensor:
        x = self.self_attn(
            x,
            x,
            x,
            attn_mask=attn_mask,
            key_padding_mask=key_padding_mask,
            need_weights=False,
            is_causal=is_causal,
        )[0]
        return self.dropout1(x)

    def _feed_forward_block(self, x: Tensor) -> Tensor:
        x = self.linear2(self.dropout(self.activation(self.linear1(x))))
        return self.dropout2(x)

    @torch.no_grad()
    def reset_parameters(self, scheme="xavier_uniform") -> None:  # noqa: D102
        logger.debug(f"Resetting parameters of {self.__class__.__name__} with scheme: {scheme}")
        self.linear1.reset_parameters(scheme)
        self.linear2.reset_parameters(scheme)
        self.norm1.reset_parameters()
        self.norm2.reset_parameters()
        # Reset attention layers:
        self.self_attn.reset_parameters(scheme)


class eTransformerDecoderLayer(torch.nn.Module):
    r"""Equivariant Transformer decoder layer mirroring :class:`torch.nn.TransformerDecoderLayer`.

    Combines an equivariant self-attention block, an equivariant cross-attention block,
    and the same eLinear/eLayerNorm feed-forward structure used by the encoder so every
    submodule commutes with the group action while keeping PyTorch’s runtime logic intact.

    The layer defines:

    .. math::
        \mathbf{f}_{\mathbf{\theta}}: \mathcal{X}^{T_{\mathrm{tgt}}}\times\mathcal{X}^{T_{\mathrm{mem}}}
        \to \mathcal{X}^{T_{\mathrm{tgt}}}.

    Functional equivariance constraint (assuming ``tgt`` and ``memory`` transform under the same representation):

    .. math::
        \mathbf{f}_{\mathbf{\theta}}(\rho_{\mathcal{X}}(g)\mathbf{tgt},\rho_{\mathcal{X}}(g)\mathbf{mem})
        = \rho_{\mathcal{X}}(g)\,\mathbf{f}_{\mathbf{\theta}}(\mathbf{tgt},\mathbf{mem})
        \quad \forall g\in\mathbb{G}.
    """

    __constants__ = ["norm_first"]

    def __init__(
        self,
        in_rep: Representation,
        nhead: int,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: Union[str, Callable[[Tensor], Tensor]] = F.relu,
        layer_norm_eps: float = 1e-5,
        batch_first: bool = True,
        norm_first: bool = True,
        norm_module: Literal["layernorm", "rmsnorm"] = "rmsnorm",
        bias: bool = True,
        device=None,
        dtype=None,
        init_scheme: str | None = "xavier_uniform",
    ) -> None:
        r"""Create an equivariant Transformer decoder layer.

        Args:
            in_rep (:class:`~escnn.group.Representation`): Input representation :math:`\rho_{\text{in}}`.
            nhead: Number of attention heads.
            dim_feedforward: Hidden dimension of the feedforward network.
            dropout: Dropout probability.
            activation: Activation function (``'relu'`` or ``'gelu'``).
            layer_norm_eps: Epsilon for layer normalization.
            batch_first: If ``True``, input/output shape is ``(B, T, D)``.
            norm_first: If ``True``, apply normalization before attention/feedforward.
            norm_module: Normalization layer type (``'layernorm'`` or ``'rmsnorm'``).
            bias: Whether to use bias in linear layers.
            device: Tensor device.
            dtype: Tensor dtype.
            init_scheme: Initialization scheme for equivariant layers.
        """
        super().__init__()
        if dim_feedforward <= 0:
            raise ValueError(f"dim_feedforward must be positive, got {dim_feedforward}")

        self.in_rep, self.out_rep = in_rep, in_rep
        factory_kwargs = {"device": device, "dtype": dtype or torch.get_default_dtype()}

        G = in_rep.group
        num_hidden_reps = max(1, ceil(dim_feedforward / G.order()))
        self.embedding_rep = direct_sum([G.regular_representation] * num_hidden_reps)
        self.hidden_dim = self.embedding_rep.size
        self.requested_dim_feedforward = dim_feedforward

        self.self_attn = eMultiheadAttention(
            in_rep=self.in_rep,
            num_heads=nhead,
            dropout=dropout,
            bias=bias,
            batch_first=batch_first,
            device=device,
            dtype=dtype,
            init_scheme=init_scheme,
        )
        self.cross_attn = eMultiheadAttention(
            in_rep=self.in_rep,
            num_heads=nhead,
            dropout=dropout,
            bias=bias,
            batch_first=batch_first,
            device=device,
            dtype=dtype,
            init_scheme=init_scheme,
        )

        self.linear1 = eLinear(self.in_rep, self.embedding_rep, bias, init_scheme=init_scheme).to(**factory_kwargs)
        self.dropout = torch.nn.Dropout(dropout)
        self.linear2 = eLinear(self.embedding_rep, self.out_rep, bias, init_scheme=init_scheme).to(**factory_kwargs)

        self.norm_first = norm_first
        if norm_module == "layernorm":
            norm_cls = eLayerNorm
            norm_kwargs = {"bias": bias} | factory_kwargs
        elif norm_module == "rmsnorm":
            norm_cls = eRMSNorm
            norm_kwargs = factory_kwargs
        else:
            raise ValueError(f"norm_module must be 'layernorm' or 'rmsnorm', got {norm_module}")
        self.norm1 = norm_cls(self.in_rep, eps=layer_norm_eps, equiv_affine=True, **norm_kwargs)
        self.norm2 = norm_cls(self.in_rep, eps=layer_norm_eps, equiv_affine=True, **norm_kwargs)
        self.norm3 = norm_cls(self.out_rep, eps=layer_norm_eps, equiv_affine=True, **norm_kwargs)

        self.dropout1 = torch.nn.Dropout(dropout)
        self.dropout2 = torch.nn.Dropout(dropout)
        self.dropout3 = torch.nn.Dropout(dropout)

        if isinstance(activation, str):
            activation = _get_activation_fn(activation)
        self.activation = activation
        self.batch_first = batch_first

        if init_scheme is not None:
            self.reset_parameters(scheme=init_scheme)

    def forward(
        self,
        tgt: Tensor,
        memory: Tensor,
        tgt_mask: Optional[Tensor] = None,
        memory_mask: Optional[Tensor] = None,
        tgt_key_padding_mask: Optional[Tensor] = None,
        memory_key_padding_mask: Optional[Tensor] = None,
        tgt_is_causal: bool = False,
        memory_is_causal: bool = False,
    ) -> Tensor:
        r"""Pass the input through the equivariant decoder layer.

        Args:
            tgt: target/query tensor of shape ``(T, B, D)`` or ``(B, T, D)`` matching
                ``batch_first``. The last dimension must equal ``in_rep.size``.
            memory: encoder memory tensor of shape ``(S, B, D)`` or ``(B, S, D)``
                (same ``batch_first``). We assume this tensor transforms under the
                *same representation* as ``tgt``; i.e., it is typically the output
                of an equivariant encoder with representation ``in_rep``.
            tgt_mask: optional target attention mask (same semantics as PyTorch’s API).
            memory_mask: optional memory attention mask.
            tgt_key_padding_mask: optional padding mask for the target batch.
            memory_key_padding_mask: optional padding mask for the memory batch.
            tgt_is_causal: if ``True``, applies a causal mask to the target self-attention.
            memory_is_causal: if ``True``, applies a causal mask to the cross-attention.
        """
        tgt_key_padding_mask = F._canonical_mask(
            mask=tgt_key_padding_mask,
            mask_name="tgt_key_padding_mask",
            other_type=F._none_or_dtype(tgt_mask),
            other_name="tgt_mask",
            target_type=tgt.dtype,
        )
        tgt_mask = F._canonical_mask(
            mask=tgt_mask,
            mask_name="tgt_mask",
            other_type=None,
            other_name="",
            target_type=tgt.dtype,
            check_other=False,
        )

        memory_key_padding_mask = F._canonical_mask(
            mask=memory_key_padding_mask,
            mask_name="memory_key_padding_mask",
            other_type=F._none_or_dtype(memory_mask),
            other_name="memory_mask",
            target_type=memory.dtype,
        )
        memory_mask = F._canonical_mask(
            mask=memory_mask,
            mask_name="memory_mask",
            other_type=None,
            other_name="",
            target_type=memory.dtype,
            check_other=False,
        )

        x = tgt
        if self.norm_first:
            x = x + self._self_attention_block(self.norm1(x), tgt_mask, tgt_key_padding_mask, tgt_is_causal)
            x = x + self._multihead_attention_block(
                self.norm2(x), memory, memory_mask, memory_key_padding_mask, memory_is_causal
            )
            x = x + self._feed_forward_block(self.norm3(x))
        else:
            x = self.norm1(x + self._self_attention_block(x, tgt_mask, tgt_key_padding_mask, tgt_is_causal))
            x = self.norm2(
                x + self._multihead_attention_block(x, memory, memory_mask, memory_key_padding_mask, memory_is_causal)
            )
            x = self.norm3(x + self._feed_forward_block(x))

        return x

    def _self_attention_block(
        self,
        x: Tensor,
        attn_mask: Optional[Tensor] = None,
        key_padding_mask: Optional[Tensor] = None,
        is_causal: bool = False,
    ) -> Tensor:
        x = self.self_attn(
            x,
            x,
            x,
            attn_mask=attn_mask,
            key_padding_mask=key_padding_mask,
            need_weights=False,
            is_causal=is_causal,
        )[0]
        return self.dropout1(x)

    def _multihead_attention_block(
        self,
        x: Tensor,
        mem: Tensor,
        attn_mask: Optional[Tensor] = None,
        key_padding_mask: Optional[Tensor] = None,
        is_causal: bool = False,
    ) -> Tensor:
        x = self.cross_attn(
            x,
            mem,
            mem,
            attn_mask=attn_mask,
            key_padding_mask=key_padding_mask,
            need_weights=False,
            is_causal=is_causal,
        )[0]
        return self.dropout2(x)

    def _feed_forward_block(self, x: Tensor) -> Tensor:
        x = self.linear2(self.dropout(self.activation(self.linear1(x))))
        return self.dropout3(x)

    @torch.no_grad()
    def reset_parameters(self, scheme="xavier_uniform") -> None:  # noqa: D102
        logger.debug(f"Resetting parameters of {self.__class__.__name__} with scheme: {scheme}")
        # Reset equivariant linear layers (symm_learning.nn.eLinear)
        self.linear1.reset_parameters(scheme)
        self.linear2.reset_parameters(scheme)
        self.norm1.reset_parameters()
        self.norm2.reset_parameters()
        self.norm3.reset_parameters()
        # Reset attention layers:
        self.self_attn.reset_parameters(scheme)
        self.cross_attn.reset_parameters(scheme)

    @torch.no_grad()
    def check_equivariance(
        self,
        batch_size: int = 4,
        tgt_len: int = 3,
        mem_len: int = 5,
        samples: int = 20,
        atol: float = 1e-4,
        rtol: float = 1e-4,
    ) -> None:
        """Quick sanity check ensuring both attention blocks and the full layer are equivariant."""
        G = self.in_rep.group
        device = next(self.parameters()).device

        def act(rep: Representation, g, tensor: Tensor) -> Tensor:
            mat = torch.tensor(rep(g), dtype=tensor.dtype, device=device)
            return torch.einsum("ij,...j->...i", mat, tensor)

        for _ in range(samples):
            g = G.sample()
            tgt = torch.randn(batch_size, tgt_len, self.in_rep.size, device=device)
            mem = torch.randn(batch_size, mem_len, self.in_rep.size, device=device)
            g_tgt = act(self.in_rep, g, tgt)
            g_mem = act(self.in_rep, g, mem)

            sa = self._self_attention_block(tgt)
            g_sa = act(self.in_rep, g, sa)
            g_sa_exp = self._self_attention_block(g_tgt)
            assert torch.allclose(g_sa, g_sa_exp, atol=atol, rtol=rtol), (
                f"Self-attention equivarinace failed max error: {torch.max(g_sa - g_sa_exp).item():.3e}"
            )

            ca = self._multihead_attention_block(tgt, mem)
            g_ca = act(self.in_rep, g, ca)
            g_ca_exp = self._multihead_attention_block(g_tgt, g_mem)
            assert torch.allclose(g_ca, g_ca_exp, atol=atol, rtol=rtol), (
                f"Cross-attention equivarinace failed max error: {torch.max(g_ca - g_ca_exp).item():.3e}"
            )

            out = self(tgt, mem)
            g_out = act(self.in_rep, g, out)
            g_out_exp = self(g_tgt, g_mem)
            assert torch.allclose(g_out, g_out_exp, atol=atol, rtol=rtol), (
                f"Transormer decoder equivarinace failed max error: {torch.max(g_ca - g_ca_exp).item():.3e}"
            )


def _get_activation_fn(activation: str) -> Callable[[Tensor], Tensor]:
    if activation == "relu":
        return F.relu
    if activation == "gelu":
        return F.gelu
    raise RuntimeError(f"activation should be relu/gelu, not {activation}")


if __name__ == "__main__":
    import logging
    import sys
    import types
    from pathlib import Path

    # logging.basicConfig(level=logging.DEBUG)
    from escnn.group import CyclicGroup, DihedralGroup, Icosahedral

    repo_root = Path(__file__).resolve().parents[3]
    test_dir = repo_root / "test"
    sys.path.insert(0, str(repo_root))
    test_pkg = sys.modules.get("test")
    test_paths = [str(path) for path in getattr(test_pkg, "__path__", [])] if test_pkg else []
    if str(test_dir) not in test_paths:
        test_pkg = types.ModuleType("test")
        test_pkg.__path__ = [str(test_dir)]
        sys.modules["test"] = test_pkg

    from symm_learning.models.transformer.etransformer import eTransformerEncoderLayer
    from symm_learning.utils import (
        bytes_to_mb,
        check_equivariance,
        describe_memory,
        module_device_memory,
        module_memory,
    )
    from test.utils import benchmark, benchmark_eval_forward

    G = CyclicGroup(2)
    m = 2
    in_rep = direct_sum([G.regular_representation] * m)

    encoder_kwargs = dict(
        in_rep=in_rep,
        nhead=1,
        dim_feedforward=in_rep.size * 4,
        dropout=0.1,
        activation="relu",
        norm_first=True,
        batch_first=True,
    )
    etransformer = eTransformerEncoderLayer(**encoder_kwargs)
    etransformer.eval()  # disable dropout for the test
    # describe_memory("transformer encoder", etransformer)
    check_equivariance(
        lambda x: etransformer._feed_forward_block(x),
        in_rep=etransformer.in_rep,
        out_rep=etransformer.out_rep,
        module_name="feed forward",
    )
    check_equivariance(
        lambda x: etransformer._self_attention_block(x),
        in_rep=etransformer.in_rep,
        out_rep=etransformer.out_rep,
        module_name="self_attention",
    )

    for depth in [1, 3, 5, 10]:
        base_layer = eTransformerEncoderLayer(**encoder_kwargs)
        base_layer.reset_parameters()
        base_layer.eval()
        encoder_stack = torch.nn.TransformerEncoder(
            encoder_layer=base_layer, num_layers=depth, enable_nested_tensor=False
        )
        for layer in encoder_stack.layers:
            if hasattr(layer, "reset_parameters"):
                layer.reset_parameters()
        encoder_stack.eval()
        print(f"\n Testing encoder stack depth={depth} equivariance...")
        check_equivariance(
            encoder_stack,
            input_dim=3,
            module_name=f"encoder stack depth={depth}",
            atol=1e-4,
            rtol=1e-4,
            in_rep=in_rep,
            out_rep=in_rep,
        )
        print(f"Encoder stack depth={depth} equivariance test passed")

    print("\n\n\nTesting decoder layer equivariance...")

    decoder_kwargs = dict(
        in_rep=in_rep,
        nhead=1,
        dim_feedforward=in_rep.size * 2,
        dropout=0.0,
        activation="relu",
        norm_first=True,
        batch_first=True,
    )
    tdecoder = eTransformerDecoderLayer(**decoder_kwargs)
    tdecoder.eval()
    tdecoder.check_equivariance()

    def check_decoder_stack(module: torch.nn.Module, rep: Representation, depth: int, atol=1e-4, rtol=1e-4):  # noqa: D103
        G = rep.group

        def act(rep: Representation, g, tensor: Tensor) -> Tensor:
            mat = torch.tensor(rep(g), dtype=tensor.dtype, device=tensor.device)
            return torch.einsum("ij,...j->...i", mat, tensor)

        B, tgt_len, mem_len = 11, 3, 5
        module.eval()
        for _ in range(10):
            g = G.sample()
            tgt = torch.randn(B, tgt_len, rep.size)
            mem = torch.randn(B, mem_len, rep.size)
            out = module(tgt=tgt, memory=mem)
            g_tgt = act(rep, g, tgt)
            g_mem = act(rep, g, mem)
            g_out = module(tgt=g_tgt, memory=g_mem)
            g_out_exp = act(rep, g, out)
            assert torch.allclose(g_out, g_out_exp, atol=atol, rtol=rtol), (
                f"Decoder stack depth={depth} equivariance failed, max err {(g_out - g_out_exp).abs().max().item():.3e}"
            )

    for depth in (1, 3, 5, 10):
        base_layer = eTransformerDecoderLayer(**decoder_kwargs)
        base_layer.reset_parameters()
        base_layer.eval()
        decoder_stack = torch.nn.TransformerDecoder(decoder_layer=base_layer, num_layers=depth)
        for layer in decoder_stack.layers:
            if hasattr(layer, "reset_parameters"):
                layer.reset_parameters()
        decoder_stack.eval()
        print(f"\n Testing decoder stack depth={depth} equivariance...")
        check_decoder_stack(decoder_stack, in_rep, depth=depth, atol=1e-3, rtol=1e-3)
        print(f"Decoder stack depth={depth} equivariance test passed")
    print("Decoder equivariance test passed")

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    torch.set_float32_matmul_precision("high")
    print(f"\nBenchmarking on device: {device}")

    fastpath_prev = torch.backends.mha.get_fastpath_enabled()
    torch.backends.mha.set_fastpath_enabled(False)

    batch_size = 256
    src_len = 16
    tgt_len = 16
    mem_len = 16
    iters = 200
    warmup = 50

    requested_dim_feedforward = in_rep.size * 4
    effective_dim_feedforward = max(1, ceil(requested_dim_feedforward / in_rep.group.order())) * in_rep.group.order()

    bench_encoder_kwargs = dict(
        in_rep=in_rep,
        nhead=1,
        dim_feedforward=requested_dim_feedforward,
        dropout=0.1,
        activation="relu",
        norm_first=True,
        batch_first=True,
        norm_module="rmsnorm",
        bias=True,
    )
    bench_decoder_kwargs = dict(
        in_rep=in_rep,
        nhead=1,
        dim_feedforward=requested_dim_feedforward,
        dropout=0.1,
        activation="relu",
        norm_first=True,
        batch_first=True,
        norm_module="rmsnorm",
        bias=True,
    )

    src = torch.randn(batch_size, src_len, in_rep.size, device=device)
    tgt = torch.randn(batch_size, tgt_len, in_rep.size, device=device)
    mem = torch.randn(batch_size, mem_len, in_rep.size, device=device)

    encoder_modules = [
        ("eTransformer Encoder", eTransformerEncoderLayer(**bench_encoder_kwargs).to(device)),
        (
            "Torch Encoder",
            torch.nn.TransformerEncoderLayer(
                d_model=in_rep.size,
                nhead=bench_encoder_kwargs["nhead"],
                dim_feedforward=effective_dim_feedforward,
                dropout=bench_encoder_kwargs["dropout"],
                activation=bench_encoder_kwargs["activation"],
                batch_first=bench_encoder_kwargs["batch_first"],
                norm_first=bench_encoder_kwargs["norm_first"],
                bias=bench_encoder_kwargs["bias"],
            ).to(device),
        ),
    ]
    decoder_modules = [
        ("eTransformer Decoder", eTransformerDecoderLayer(**bench_decoder_kwargs).to(device)),
        (
            "Torch Decoder",
            torch.nn.TransformerDecoderLayer(
                d_model=in_rep.size,
                nhead=bench_decoder_kwargs["nhead"],
                dim_feedforward=effective_dim_feedforward,
                dropout=bench_decoder_kwargs["dropout"],
                activation=bench_decoder_kwargs["activation"],
                batch_first=bench_decoder_kwargs["batch_first"],
                norm_first=bench_decoder_kwargs["norm_first"],
                bias=bench_decoder_kwargs["bias"],
            ).to(device),
        ),
    ]

    def print_benchmark_table(title: str, results: list[dict], name_width: int) -> None:  # noqa: D103
        header = (
            f"{'Layer':<{name_width}} {'Forward eval (ms)':>18} {'Forward (ms)':>18} {'Backward (ms)':>18} "
            f"{'Total (ms)':>15} {'Trainable MB':>15} {'Non-train MB':>15} {'Total MB':>12} "
            f"{'GPU Alloc MB':>15} {'GPU Peak MB':>15}"
        )
        separator = "-" * len(header)
        print(f"\n{title}")
        print(separator)
        print(header)
        print(separator)
        for res in results:
            fwd_eval_str = f"{res['fwd_eval_mean']:.3f} +/- {res['fwd_eval_std']:.3f}"
            fwd_str = f"{res['fwd_mean']:.3f} +/- {res['fwd_std']:.3f}"
            bwd_str = f"{res['bwd_mean']:.3f} +/- {res['bwd_std']:.3f}"
            total_mb = res["train_mem"] + res["non_train_mem"]
            gpu_alloc_mb = bytes_to_mb(res["gpu_mem"])
            gpu_peak_mb = bytes_to_mb(res["gpu_peak"])
            print(
                f"{res['name']:<{name_width}} {fwd_eval_str:>18} {fwd_str:>18} {bwd_str:>18} "
                f"{res['total_time']:>15.3f} {bytes_to_mb(res['train_mem']):>15.3f} "
                f"{bytes_to_mb(res['non_train_mem']):>15.3f} {bytes_to_mb(total_mb):>12.3f} "
                f"{gpu_alloc_mb:>15.3f} {gpu_peak_mb:>15.3f}"
            )
        print(separator)

    def benchmark_modules(  # noqa: D103
        title: str,
        modules: list[tuple[str, torch.nn.Module]],
        forward_fn_builder: Callable[[torch.nn.Module], torch.Tensor],
    ) -> None:
        results = []
        for name, module in modules:

            def forward_fn(mod=module):  # noqa: D103
                return forward_fn_builder(mod)

            train_mem, non_train_mem = module_memory(module)
            gpu_alloc, gpu_peak = module_device_memory(module, device=device)
            eval_fwd_mean, eval_fwd_std = benchmark_eval_forward(module, forward_fn, iters=iters, warmup=warmup)
            (fwd_mean, fwd_std), (bwd_mean, bwd_std) = benchmark(module, forward_fn, iters=iters, warmup=warmup)
            results.append(
                {
                    "name": name,
                    "fwd_eval_mean": eval_fwd_mean,
                    "fwd_eval_std": eval_fwd_std,
                    "fwd_mean": fwd_mean,
                    "fwd_std": fwd_std,
                    "bwd_mean": bwd_mean,
                    "bwd_std": bwd_std,
                    "total_time": fwd_mean + bwd_mean,
                    "train_mem": train_mem,
                    "non_train_mem": non_train_mem,
                    "gpu_mem": gpu_alloc,
                    "gpu_peak": gpu_peak,
                }
            )

        name_width = max(22, max(len(res["name"]) for res in results) + 2)
        print_benchmark_table(title, results, name_width)

    benchmark_modules(
        title=f"Encoder layer benchmark per batch={batch_size}, seq_len={src_len}",
        modules=encoder_modules,
        forward_fn_builder=lambda mod: mod(src),
    )
    benchmark_modules(
        title=f"Decoder layer benchmark per batch={batch_size}, tgt_len={tgt_len}, mem_len={mem_len}",
        modules=decoder_modules,
        forward_fn_builder=lambda mod: mod(tgt, mem),
    )

    torch.backends.mha.set_fastpath_enabled(fastpath_prev)



# Created by Daniel Ordoñez (daniels.ordonez@gmail.com) at 02/04/25
from __future__ import annotations

from collections.abc import Callable

import numpy as np
import torch
from escnn.group import Representation


class CallableDict(dict, Callable):
    """Dictionary that can be called as a function."""

    def __call__(self, key):
        """Return the value of the key."""
        return self[key]


def _tensor_bytes(t: torch.Tensor) -> int:
    if t.layout == torch.sparse_coo:
        vals = t._values()
        idx = t._indices()
        return vals.numel() * vals.element_size() + idx.numel() * idx.element_size()
    if t.layout == torch.sparse_csr or t.layout == torch.sparse_bsr:
        total = t.values().numel() * t.values().element_size()
        total += t.crow_indices().numel() * t.crow_indices().element_size()
        total += t.col_indices().numel() * t.col_indices().element_size()
        return total
    if t.layout == torch.sparse_csc:
        total = t.values().numel() * t.values().element_size()
        total += t.ccol_indices().numel() * t.ccol_indices().element_size()
        total += t.row_indices().numel() * t.row_indices().element_size()
        return total
    return t.numel() * t.element_size()


def module_memory(module: torch.nn.Module, units: str = "bytes"):
    """Return trainable/non-trainable parameter memory for ``module``.

    Args:
        module (:class:`torch.nn.Module`): Module whose parameter memory footprint is inspected.
        units (:class:`str`, optional): Output units. One of ``"bytes"``, ``"kib"``, ``"mib"``, ``"gib"``.
             Default is raw bytes.

    Returns:
        :class:`tuple` [:class:`float`, :class:`float`]: Memory in the selected units stored in parameters that
            require gradients, and frozen parameters plus buffers.
    """
    unit_scale = {
        "bytes": 1,
        "kib": 1024,
        "mib": 1024**2,
        "gib": 1024**3,
    }
    if units.lower() not in unit_scale:
        raise ValueError(f"Unsupported units '{units}'. Expected one of {tuple(unit_scale.keys())}.")
    scale = unit_scale[units.lower()]

    trainable = 0
    frozen = 0
    for p in module.parameters():
        nbytes = _tensor_bytes(p)
        if p.requires_grad:
            trainable += nbytes
        else:
            frozen += nbytes
    buffer_bytes = sum(_tensor_bytes(buf) for buf in module.buffers())
    non_trainable = frozen + buffer_bytes
    return trainable / scale, non_trainable / scale


def backprop_sanity(module: torch.nn.Module) -> None:
    """Simple backpropagation sanity check for ``module``."""
    module.train()
    optim = torch.optim.SGD(module.parameters(), lr=1e-3)
    x = torch.randn(16, module.in_rep.size)
    target = torch.randn(16, module.out_rep.size)
    optim.zero_grad()
    y = module(x)
    loss = torch.nn.functional.mse_loss(y, target)
    loss.backward()
    grad_norms = [p.grad.norm().item() for p in module.parameters() if p.grad is not None]
    assert grad_norms, "Expected at least one gradient to propagate."
    optim.step()


def module_memory_breakdown(module: torch.nn.Module) -> list[dict]:
    """Return a per-tensor memory breakdown for ``module`` (sizes in bytes)."""

    def _entry(name: str, tensor: torch.Tensor, kind: str, trainable: bool) -> dict:
        return {
            "name": name,
            "kind": kind,
            "trainable": bool(trainable),
            "shape": tuple(tensor.shape),
            "dtype": str(tensor.dtype).replace("torch.", ""),
            "numel": tensor.numel(),
            "bytes": _tensor_bytes(tensor),
        }

    details: list[dict] = []
    for name, param in module.named_parameters():
        details.append(_entry(name, param, "parameter", param.requires_grad))
    for name, buf in module.named_buffers():
        details.append(_entry(name, buf, "buffer", False))

    return details


def describe_memory(label: str, module: torch.nn.Module) -> None:
    """Pretty-print parameter/buffer memory consumption for ``module``."""
    entries = module_memory_breakdown(module)
    if not entries:
        print(f"\n{label}: no parameters or buffers.")
        return
    entries.sort(key=lambda item: item["bytes"], reverse=True)

    print(f"\n{label} memory breakdown")
    header = f"{'Kind':<12}{'Name':<45}{'Trainable':<11}{'Shape':<25}{'DType':<12}{'MB':>10}"
    print(header)
    print("-" * len(header))
    for entry in entries:
        shape = "×".join(str(dim) for dim in entry["shape"]) if entry["shape"] else "scalar"
        mb = bytes_to_mb(entry["bytes"])
        print(
            f"{entry['kind']:<12}{entry['name']:<45}{str(entry['trainable']):<11}"
            f"{shape:<25}{entry['dtype']:<12}{mb:>10.3f}"
        )
    total_trainable = sum(bytes_to_mb(e["bytes"]) for e in entries if e["kind"] == "parameter" and e["trainable"])
    total_frozen = sum(bytes_to_mb(e["bytes"]) for e in entries if e["kind"] == "parameter" and not e["trainable"])
    total_buffers = sum(bytes_to_mb(e["bytes"]) for e in entries if e["kind"] == "buffer")
    print("-" * len(header))
    print(f"{'Trainable params MB:':<68}{total_trainable:>10.3f}")
    print(f"{'Frozen params MB:':<68}{total_frozen:>10.3f}")
    print(f"{'Buffers MB:':<68}{total_buffers:>10.3f}")


def module_device_memory(module: torch.nn.Module, device: str | torch.device | None = None) -> tuple[int, int]:
    """Measure CUDA memory footprint (bytes allocated/peak) of ``module`` cloned onto ``device``."""
    if device is None:
        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    device = torch.device(device)
    if device.type != "cuda":
        return 0, 0

    torch.cuda.empty_cache()
    torch.cuda.reset_peak_memory_stats(device)
    torch.cuda.synchronize(device)

    module.to("cpu")
    before = torch.cuda.memory_allocated(device)
    module_clone = module.to(device)
    torch.cuda.synchronize(device)
    allocated = torch.cuda.memory_allocated(device) - before
    peak = torch.cuda.max_memory_allocated(device)

    del module_clone
    torch.cuda.empty_cache()
    return allocated, peak


def bytes_to_mb(num_bytes: int) -> float:  # noqa: D103
    return num_bytes / (1024**2)


def get_spectral_trivial_dims(rep: Representation) -> list[int]:
    """Return the list of indices of the trivial irreps in the spectral decomposition of ``rep``."""
    G = rep.group
    trivial_id = G.trivial_representation.id
    trivial_spectral_dims = []
    offset = 0
    for irrep_id in rep.irreps:
        if irrep_id == trivial_id:
            trivial_spectral_dims.append(offset)
        offset += G.irrep(*irrep_id).size

    return trivial_spectral_dims


def get_spectral_trivial_mask(rep: Representation) -> torch.Tensor:
    """Return a boolean mask selecting the trivial irreps in the spectral decomposition of ``rep``."""
    trivial_dims = get_spectral_trivial_dims(rep)
    mask = torch.zeros(rep.size, dtype=torch.bool)
    mask[trivial_dims] = 1
    return mask


def check_equivariance(
    e_module,
    atol: float = 1e-5,
    rtol: float = 1e-5,
    input_dim: int | tuple[int] = 2,
    in_rep: Representation | tuple[Representation, ...] = None,
    out_rep: Representation | tuple[Representation, ...] = None,
    module_name: str = None,
):
    """Method that automatically tests the equivariance of the current module."""
    in_rep = e_module.in_rep if hasattr(e_module, "in_rep") else in_rep
    out_rep = e_module.out_rep if hasattr(e_module, "out_rep") else out_rep
    assert in_rep is not None, f"in_rep must be provided or be an attribute of the module {e_module}"
    assert out_rep is not None, f"out_rep must be provided or be an attribute of the module {e_module}"
    G = in_rep.group
    module_name = e_module.__class__.__name__ if module_name is None else module_name
    batch_size = 20
    spatial_dims = 9

    if input_dim == 2:
        x = torch.randn(batch_size, in_rep.size)
    elif input_dim == 3:
        x = torch.randn(batch_size, spatial_dims, in_rep.size)
    else:
        raise ValueError(f"Input dimension {input_dim} not supported.")

    dtype, device = x.dtype, x.device

    errors = []

    if isinstance(e_module, torch.nn.Module):
        e_module.eval()  # Disable dropout, batchnorm, etc.

    # for el in self.out_type.testing_elements:
    for _ in range(20):
        g = G.sample()
        gx = torch.einsum("ij,...j->...i", torch.tensor(in_rep(g), dtype=dtype, device=device), x)
        y = e_module(x)
        gy = torch.einsum("ij,...j->...i", torch.tensor(out_rep(g)).to(dtype=dtype, device=device), y)
        gy = gy.detach().cpu().numpy()

        gy_expected = e_module(gx).detach().cpu().numpy()

        errs = gy - gy_expected
        errs = np.abs(errs).reshape(-1)

        assert gy.shape == gy_expected.shape, f"Shape mismatch: got {gy.shape}, expected {gy_expected.shape}"
        assert np.allclose(gy, gy_expected, atol=atol, rtol=rtol), (
            f"Equivariance test failed for group element {g} max scalar error {errs.max()}"
        )

        errors.append((g, errs.mean()))

    print(f"Equivariant check passed for module {module_name} with max error {max(e[1] for e in errors)} ")



"""Symmetric Learning - Statistics Utilities.

Functions for computing statistics of symmetric random variables that respect group
symmetry constraints.

Functions
---------
mean
    Compute the mean projected onto the G-invariant subspace.
var
    Compute variance respecting symmetry structure.
var_mean
    Compute variance and mean together efficiently.
cov
    Compute covariance between symmetric random variables.
"""

from __future__ import annotations

import numpy as np
import torch
from escnn.group import Representation
from torch import Tensor

from symm_learning.linalg import invariant_orthogonal_projector
from symm_learning.representation_theory import isotypic_decomp_rep


def mean(x: Tensor, rep_x: Representation) -> Tensor:
    r"""Estimate the :math:`\mathbb{G}`-invariant mean of a random variable.

    Let :math:`\mathbf{X}: \Omega \to \mathcal{X}` be a random variable taking values in the symmetric vector space
    :math:`\mathcal{X}`, with group representation :math:`\rho_{\mathcal{X}}:\mathbb{G}\to\mathrm{GL}(\mathcal{X})`,
    and marginal density :math:`\mathbb{P}_{\mathbf{X}}`.
    Under the assumption that this marginal is invariant under the group action
    (i.e., a point and all its symmetric points have equal likelihood under the marginal), formally:

    .. math::
        \mathbb{P}_{\mathbf{X}}(\mathbf{x})
        = \mathbb{P}_{\mathbf{X}}\!\left(\rho_{\mathcal{X}}(g)\mathbf{x}\right),
        \quad \forall \mathbf{x}\in\mathcal{X},\ \forall g\in\mathbb{G},

    the true mean satisfies

    .. math::
        \mathbb{E}[\mathbf{X}] = \rho_{\mathcal{X}}(g)\,\mathbb{E}[\mathbf{X}], \quad \forall g\in\mathbb{G},

    hence :math:`\mathbb{E}[\mathbf{X}] \in \mathcal{X}^{\text{inv}}`.

    Implementation:
    from samples :math:`\{\mathbf{x}^{(n)}\}_{n=1}^N`, we first compute the empirical mean

    .. math::
        \widehat{\mathbb{E}}[\mathbf{X}] = \frac{1}{N}\sum_{n=1}^N \mathbf{x}^{(n)},

    and then project it onto the invariant subspace:

    .. math::
        \widehat{\mathbb{E}}_{\mathbb{G}}[\mathbf{X}]
        = \mathbf{P}_{\mathrm{inv}}\,\widehat{\mathbb{E}}[\mathbf{X}],
        \quad
        \mathbf{P}_{\mathrm{inv}} = \mathbf{Q}\mathbf{S}\mathbf{Q}^T,

    where :math:`\mathbf{S}` selects trivial-irrep coordinates in the irrep-spectral basis.
    (see :func:`~symm_learning.linalg.invariant_orthogonal_projector`).
    Under the repository's canonical isotypic ordering, this corresponds to the first isotypic block when present.

    Args:
        x: (:class:`torch.Tensor`) samples of :math:`\mathbf{X}` with shape :math:`(N,D_x)` or
            :math:`(N,D_x,T)`; when a time axis is present it is folded into the sample axis.
        rep_x: (:class:`~escnn.group.Representation`) representation :math:`\rho_{\mathcal{X}}` on :math:`\mathcal{X}`.

    Returns:
        (:class:`torch.Tensor`): Invariant mean vector in :math:`\mathcal{X}`.

    Shape:
        - **x**: :math:`(N,D_x)` or :math:`(N,D_x,T)`.
        - **Output**: :math:`(D_x,)`.

    Note:
        For repeated calls with the same representation object ``rep_x``, this function caches
        :math:`\mathbf{P}_{\mathrm{inv}}` in ``rep_x.attributes["invariant_orthogonal_projector"]``.
    """
    assert x.ndim in (2, 3), f"Expected x to be a 2D or 3D tensor, got {x.ndim}D tensor"

    if "invariant_orthogonal_projector" not in rep_x.attributes:
        P_inv = invariant_orthogonal_projector(rep_x)
        rep_x.attributes["invariant_orthogonal_projector"] = P_inv
    else:
        P_inv = rep_x.attributes["invariant_orthogonal_projector"]

    x_flat = x if x.ndim == 2 else x.reshape(-1, x.shape[1])

    mean_empirical = torch.mean(x_flat, dim=0)  # Mean over batch as sequence length.
    # Project to the inv-subspace and map back to the original basis
    mean_result = torch.einsum("ij,j->i...", P_inv.to(device=x_flat.device, dtype=mean_empirical.dtype), mean_empirical)
    return mean_result


def var(x: Tensor, rep_x: Representation, center: Tensor = None) -> Tensor:
    r"""Estimate the symmetry-constrained variance of :math:`\mathbf{X}:\Omega\to\mathcal{X}`.

    Let :math:`\mathbf{X}: \Omega \to \mathcal{X}` be a random variable taking values in the symmetric vector space
    :math:`\mathcal{X}`, with group representation :math:`\rho_{\mathcal{X}}:\mathbb{G}\to\mathrm{GL}(\mathcal{X})`,
    and marginal density :math:`\mathbb{P}_{\mathbf{X}}`.
    Under the assumption that this marginal is invariant under the group action
    (i.e., a point and all its symmetric points have equal likelihood under the marginal), formally:

    .. math::
        \mathbb{P}_{\mathbf{X}}(\mathbf{x})
        = \mathbb{P}_{\mathbf{X}}\!\left(\rho_{\mathcal{X}}(g)\mathbf{x}\right),
        \quad \forall \mathbf{x}\in\mathcal{X},\ \forall g\in\mathbb{G},

    the true variance in the irrep-spectral basis
    (:func:`~symm_learning.representation_theory.isotypic_decomp_rep`) is constant within each irreducible copy:

    .. math::
        \operatorname{Var}(\hat{\mathbf{X}}_{k,i,1})
        = \cdots =
        \operatorname{Var}(\hat{\mathbf{X}}_{k,i,d_k})
        = \sigma^2_{k,i}.

    Implementation:
    given samples :math:`\{\mathbf{x}^{(n)}\}_{n=1}^{N}`, we compute:

    1. Centering (using provided center or :func:`mean`):

    .. math::
        \widehat{\boldsymbol{\mu}} =
        \begin{cases}
        \texttt{center}, & \text{if provided} \\
        \widehat{\mathbb{E}}_{\mathbb{G}}[\mathbf{X}], & \text{otherwise}
        \end{cases}

    2. Empirical spectral variance:

    .. math::
        \hat{\mathbf{x}}^{(n)} = \mathbf{Q}^{T}(\mathbf{x}^{(n)}-\widehat{\boldsymbol{\mu}}),\qquad
        \widehat{v}_{j} = \frac{1}{N-1}\sum_{n=1}^{N}\left(\hat{x}^{(n)}_{j}\right)^2.

    3. Irrep-wise averaging for each copy :math:`(k,i)`:

    .. math::
        \widehat{\sigma}^{2}_{k,i}
        = \frac{1}{d_k}\sum_{r=1}^{d_k}\widehat{v}_{k,i,r},
        \qquad
        \widehat{v}_{k,i,1}=\cdots=\widehat{v}_{k,i,d_k}:=\widehat{\sigma}^{2}_{k,i}.

    4. Mapping back to the original basis:

    .. math::
        \widehat{\operatorname{Var}}(\mathbf{X}) = \mathbf{Q}^{\odot 2}\,\widehat{\mathbf{v}},

    where :math:`\mathbf{Q}^{\odot 2}` is the elementwise square of :math:`\mathbf{Q}` and
    :math:`\widehat{\mathbf{v}}` denotes the broadcast spectral variance vector after step 3.

    Args:
        x: (:class:`torch.Tensor`) samples with shape :math:`(N,D_x)` or :math:`(N,D_x,T)`;
            the optional time axis is folded into samples.
        rep_x: (:class:`~escnn.group.Representation`) representation :math:`\rho_{\mathcal{X}}`.
        center: (:class:`torch.Tensor`, optional) Center for variance computation. If None, computes the mean.

    Returns:
        (:class:`torch.Tensor`): Variance vector in the original basis, consistent with the irrep-wise constraint above.

    Shape:
        - **x**: :math:`(N,D_x)` or :math:`(N,D_x,T)`.
        - **center**: :math:`(D_x,)` if provided.
        - **Output**: :math:`(D_x,)`.

    Note:
        For repeated calls with the same representation object ``rep_x``, this function caches and reuses:
        ``Q_inv``, ``Q_squared``, ``irrep_dims``, and ``irrep_indices`` in ``rep_x.attributes``.
    """
    assert x.ndim in (2, 3), f"Expected x to be a 2D or 3D tensor, got {x.ndim}D tensor"

    if "Q_inv" not in rep_x.attributes:  # Use cache Tensor if available.
        Q_inv = torch.tensor(rep_x.change_of_basis_inv, device=x.device, dtype=x.dtype)
        rep_x.attributes["Q_inv"] = Q_inv
    else:
        Q_inv = rep_x.attributes["Q_inv"]

    if "Q_squared" not in rep_x.attributes:  # Use cache Tensor if available.
        Q = torch.tensor(rep_x.change_of_basis, device=x.device, dtype=x.dtype)
        Q_squared = Q.pow(2)
        rep_x.attributes["Q_squared"] = Q_squared
    else:
        Q_squared = rep_x.attributes["Q_squared"]

    x_flat = x if x.ndim == 2 else x.reshape(-1, x.shape[1])

    # Use provided center or compute mean
    if center is None:
        center = mean(x, rep_x)

    # Symmetry constrained variance computation.
    # The variance is constraint to be a single constant per each irreducible subspace.
    # Hence, we compute the empirical variance, and average within each irreducible subspace.
    n_samples = x_flat.shape[0]

    x_c_irrep_spectral = torch.einsum("ij,...j->...i", Q_inv.to(device=x_flat.device), x_flat - center)
    var_spectral = torch.sum(x_c_irrep_spectral**2, dim=0) / (n_samples - 1)

    # Vectorized averaging over irreducible subspace dimensions
    if "irrep_dims" not in rep_x.attributes:
        irrep_dims = torch.tensor([rep_x.group.irrep(*irrep_id).size for irrep_id in rep_x.irreps], device=x.device)
        rep_x.attributes["irrep_dims"] = irrep_dims
    else:
        irrep_dims = rep_x.attributes["irrep_dims"].to(device=x.device)

    if "irrep_indices" not in rep_x.attributes:
        # Create indices for each irrep subspace: [0,0,0,1,1,2,2,2,2,...] for irrep dims [3,2,4,...]
        indices = torch.arange(len(irrep_dims)).to(device=irrep_dims.device)
        irrep_indices = torch.repeat_interleave(indices, irrep_dims)
        rep_x.attributes["irrep_indices"] = irrep_indices
    else:
        irrep_indices = rep_x.attributes["irrep_indices"].to(device=x.device)

    # Compute average variance for each irrep subspace using scatter operations
    avg_vars = torch.zeros(len(irrep_dims), device=x.device, dtype=var_spectral.dtype)

    # Sum variances within each irrep subspace using scatter_add_:
    # For irrep_indices = [0,0,0,1,1,2,2,2,2] and var_spectral = [v0,v1,v2,v3,v4,v5,v6,v7,v8]
    # This computes: avg_vars[0] = v0+v1+v2, avg_vars[1] = v3+v4, avg_vars[2] = v5+v6+v7+v8
    avg_vars.scatter_add_(0, irrep_indices, var_spectral)
    avg_vars = avg_vars / irrep_dims

    # Broadcast back to full spectral dimensions
    var_spectral = avg_vars[irrep_indices]

    var_result = torch.einsum("ij,...j->...i", Q_squared.to(device=x.device), var_spectral)
    return var_result


def var_mean(x: Tensor, rep_x: Representation):
    r"""Compute :func:`var` and :func:`mean` under symmetry constraints.

    Args:
        x: (:class:`torch.Tensor`) samples with shape :math:`(N,D_x)` or :math:`(N,D_x,T)`.
        rep_x: (:class:`~escnn.group.Representation`) representation :math:`\rho_{\mathcal{X}}`.

    Returns:
        (:class:`torch.Tensor`, :class:`torch.Tensor`): Tuple ``(var, mean)`` where mean is projected to
        :math:`\mathcal{X}^{\text{inv}}` and variance satisfies irrep-wise isotropy in spectral basis.

    Shape:
        - **x**: :math:`(N,D_x)` or :math:`(N,D_x,T)`.
        - **Output**: ``(var, mean)`` both with shape :math:`(D_x,)`.

    Note:
        This function reuses the same caches as :func:`mean` and :func:`var` when called repeatedly with the
        same representation object ``rep_x``.
    """
    # Compute mean first
    mean_result = mean(x, rep_x)
    # Compute variance using the computed mean
    var_result = var(x, rep_x, center=mean_result)
    return var_result, mean_result


def _isotypic_cov(x: Tensor, rep_x: Representation, y: Tensor = None, rep_y: Representation = None):
    r"""Cross-covariance between two **isotypic sub-spaces that share the same irrep**.

    If both signals live in
    :math:`\rho_X=\bigoplus_{i=1}^{m_x}\rho_k` and
    :math:`\rho_Y=\bigoplus_{i=1}^{m_y}\rho_k`
    (with :math:`\rho_k` of dimension *d*), every
    :math:`G`-equivariant linear map factorises as

    .. math::
        \operatorname{Cov}(\mathbf{X}, \mathbf{Y})
        \;=\;\mathbf{Z}_{XY}\otimes \mathbf{I}_d,  \qquad
        \mathbf{Z}_{XY}\in\mathbb{R}^{m_y\times m_x}.

    We estimate the free matrix :math:`\mathbf Z_{XY}` by

    1. **centering** (skipped if the irrep is trivial);
    2. **reshaping** the data so that each copy of the irrep becomes one
       “channel” of length *d·N*;
    3. **projecting** every :math:`d\times d` block onto the orthogonal basis
       of :math:`\mathrm{End}_{\mathbb{G}}(\rho_k)` via Frobenius inner products (see
       `arXiv:2505.19809 <https://arxiv.org/abs/2505.19809>`_);
    4. rebuilding the block matrix that respects the constraint above.

    When ``y is None`` the routine reduces to an **auto-covariance** and only
    the symmetric (identity) basis element is kept.


    Args:
        x (:class:`torch.Tensor`): shape :math:`(N,\; m_x d)` — samples drawn from ``rep_x``.
        rep_x (:class:`~escnn.group.Representation`): isotypic representation
            containing exactly one irrep type.
        y (:class:`torch.Tensor`, optional): shape :math:`(N,\; m_y d)` —
            samples drawn from ``rep_y``.  If *None*, computes the
            auto-covariance of *x*.
        rep_y (:class:`~escnn.group.Representation`, optional): isotypic
            representation matching the irrep of ``rep_x``; ignored when
            *y* is *None*.

    Returns:
        (Tensor, Tensor):
            - C_xy: :math:`(m_y d,\; m_x d)` projected covariance.
            - Z_xy: :math:`(m_y,\; m_x,\; B)`, free coefficients of each cross-covariance between irrep subspaces,
              representing basis expansion coefficients in the basis of endomorphisms of the irrep subspaces.
              Where :math:`B = 1, 2, 4` for real, complex, quaternionic irreps, respectively.
    """
    irrep_id = rep_x.irreps[0]  # Irrep id of the isotypic subspace
    assert rep_x.size == x.shape[-1], f"Expected signal shape to be (..., {rep_x.size}) got {x.shape}"
    assert rep_x.attributes.get("is_isotypic_rep", False), f"Expected rep of a single type, got {rep_x.irreps}"

    if y is not None:
        assert rep_y.size == y.shape[-1], f"Expected signal shape to be (..., {rep_y.size}) got {y.shape}"
        assert rep_y.attributes.get("is_isotypic_rep", False), f"Expected rep of a single type, got {rep_y.irreps}"

    # Get information about the irreducible representation present in the isotypic subspace
    irrep_dim = rep_x.attributes["irrep_dim"]
    # irrep_end_basis := (dim(End(irrep)), dim(irrep), dim(irrep))
    irrep_end_basis = rep_x.attributes["irrep_endomorphism_basis"].to(device=x.device, dtype=x.dtype)

    if y is None:
        rep_y = rep_x  # Use the same representation for Y
        y = x

    m_x = rep_x.attributes["irrep_multiplicity"]  # Multiplicity of the irrep in X
    m_y = rep_y.attributes["irrep_multiplicity"]  # Multiplicity of the irrep in Y

    x_iso, y_iso = x, y

    is_inv_subspace = irrep_id == rep_x.group.trivial_representation.id
    if is_inv_subspace:  # Nothing to do, return empirical covariance.
        x_iso = x - torch.mean(x, dim=0, keepdim=True)
        y_iso = y - torch.mean(y, dim=0, keepdim=True)
        Cxy_iso = torch.einsum("...y,...x->yx", y_iso, x_iso) / (x_iso.shape[0] - 1)
        return Cxy_iso, Cxy_iso  # Invariant subspace covariance is the same as the covariance matrix.

    # Compute empirical cross-covariance
    Cxy_iso = torch.einsum("...y,...x->yx", y_iso, x_iso) / (x_iso.shape[0] - 1)
    # Reshape from (my * d, mx * d) to (my, mx, d, d)
    Cxy_irreps = Cxy_iso.view(m_y, irrep_dim, m_x, irrep_dim).permute(0, 2, 1, 3).contiguous()
    # Compute basis expansion coefficients of each irrep cross-covariance in basis of End(irrep) ========
    # Frobenius inner product  <C , Ψ_b>  =  Σ_{i,j} C_{ij} Ψ_b,ij
    Cxy_irreps_basis_coeff = torch.einsum("mnij,bij->mnb", Cxy_irreps, irrep_end_basis)  # (m_y , m_x , B)
    # squared norms ‖Ψ_b‖²
    basis_coeff_norms = torch.einsum("bij,bij->b", irrep_end_basis, irrep_end_basis)  # (B,)
    Cxy_irreps_basis_coeff = Cxy_irreps_basis_coeff / basis_coeff_norms[None, None]

    Cxy_irreps = torch.einsum("...b,bij->...ij", Cxy_irreps_basis_coeff, irrep_end_basis)  # (m_y , m_x , d , d)
    # Reshape to (my * d, mx * d)
    Cxy_iso = Cxy_irreps.permute(0, 2, 1, 3).reshape(m_y * irrep_dim, m_x * irrep_dim)

    return Cxy_iso, Cxy_irreps_basis_coeff


def cov(x: Tensor, y: Tensor, rep_x: Representation, rep_y: Representation):
    r"""Compute the covariance between two symmetric random variables.

    Let :math:`\mathbf{X}:\Omega\to\mathcal{X}` and :math:`\mathbf{Y}:\Omega\to\mathcal{Y}` with representations
    :math:`\rho_{\mathcal{X}}`, :math:`\rho_{\mathcal{Y}}`. The covariance is computed via the
    :ref:`isotypic decomposition <isotypic-decomposition-example>`
    from :func:`~symm_learning.representation_theory.isotypic_decomp_rep`.
    Covariance contributions between different isotypic types are zero in the constrained model.
    Hence, in the disentangled/isotypic basis the covariance can be computed in block-diagonal form:

    .. math::
        \begin{align}
            \mathbf{C}_{xy} &= \mathbf{Q}_y^T (\bigoplus_{k} \mathbf{C}_{xy}^{(k)} )\mathbf{Q}_x
            \\
            &= \mathbf{Q}_y^T (
            \bigoplus_{k} \sum_{b\in \mathbb{B}_k} \mathbf{Z}_b^{(k)} \otimes \mathbf{b}
            ) \mathbf{Q}_x
            \\
        \end{align}

    Where :math:`\mathbf{Q}_x^{\mathsf T}` and :math:`\mathbf{Q}_y^{\mathsf T}`
    are the change-of-basis matrices to the isotypic bases of :math:`\mathcal{X}` and :math:`\mathcal{Y}`,
    respectively; :math:`\mathbf{C}_{xy}^{(k)}` is the covariance restricted to the
    isotypic subspaces of type *k*; and :math:`\mathbf{Z}_b^{(k)}` are the free
    parameters—i.e. the expansion coefficients in the endomorphism basis
    :math:`\mathbb{B}_k` of the irreducible representation of type *k*.

    Args:
        x (:class:`torch.Tensor`): Realizations of a random variable :math:`X`.
        y (:class:`torch.Tensor`): Realizations of a random variable :math:`Y`.
        rep_x (:class:`~escnn.group.Representation`): Representation :math:`\rho_{\mathcal{X}}`.
        rep_y (:class:`~escnn.group.Representation`): Representation :math:`\rho_{\mathcal{Y}}`.

    Returns:
        :class:`torch.Tensor`: The covariance matrix between the two random variables, of shape :math:`(D_y, D_x)`.

    Shape:
        - **x**: :math:`(N, D_x)` where :math:`D_x` is the dimension of the random variable :math:`\mathbf{X}`.
        - **y**: :math:`(N, D_y)` where :math:`D_y` is the dimension of the random variable :math:`\mathbf{Y}`.

        - **Output**: :math:`(D_y, D_x)`

    Note:
        This function calls :func:`~symm_learning.representation_theory.isotypic_decomp_rep`, which caches
        decompositions in the group representation registry. Repeated calls with the same representations reuse
        cached decompositions.
    """
    # assert X.shape[0] == Y.shape[0], "Expected equal number of samples in X and Y"
    assert x.shape[1] == rep_x.size, f"Expected X shape (N, {rep_x.size}), got {x.shape}"
    assert y.shape[1] == rep_y.size, f"Expected Y shape (N, {rep_y.size}), got {y.shape}"
    assert x.shape[-1] == rep_x.size, f"Expected X shape (..., {rep_x.size}), got {x.shape}"
    assert y.shape[-1] == rep_y.size, f"Expected Y shape (..., {rep_y.size}), got {y.shape}"

    rep_X_iso = isotypic_decomp_rep(rep_x)
    rep_Y_iso = isotypic_decomp_rep(rep_y)
    rep_X_iso_subspaces = rep_X_iso.attributes["isotypic_reps"]
    rep_Y_iso_subspaces = rep_Y_iso.attributes["isotypic_reps"]
    iso_idx_X = rep_X_iso.attributes["isotypic_subspace_dims"]
    iso_idx_Y = rep_Y_iso.attributes["isotypic_subspace_dims"]

    # Changes of basis from the Disentangled/Isotypic-basis of X, and Y to the original basis.
    Qx = torch.tensor(rep_X_iso.change_of_basis, device=x.device, dtype=x.dtype)
    Qy = torch.tensor(rep_Y_iso.change_of_basis, device=y.device, dtype=y.dtype)

    X_iso = torch.einsum("ij,...j->...i", Qx.T, x)
    Y_iso = torch.einsum("ij,...j->...i", Qy.T, y)
    Cxy_iso = torch.zeros((rep_y.size, rep_x.size), dtype=x.dtype, device=x.device)
    for iso_id in rep_Y_iso_subspaces:
        if iso_id not in rep_X_iso_subspaces:
            continue  # No covariance between the isotypic subspaces of different types.
        X_k = X_iso[..., iso_idx_X[iso_id]]
        Y_k = Y_iso[..., iso_idx_Y[iso_id]]
        rep_X_k = rep_X_iso_subspaces[iso_id]
        rep_Y_k = rep_Y_iso_subspaces[iso_id]
        # Cxy_k = D_xy_k ⊗ I_d [my * d x mx * d]
        Cxy_k, _ = _isotypic_cov(x=X_k, y=Y_k, rep_x=rep_X_k, rep_y=rep_Y_k)
        Cxy_iso[iso_idx_Y[iso_id], iso_idx_X[iso_id]] = Cxy_k

    # Change to the original basis
    Cxy = Qy.T @ Cxy_iso @ Qx
    return Cxy



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from time import perf_counter
from typing import Optional, Tuple, Union

import torch

from symm_learning.models.diffusion.cond_unet1d import SinusoidalPosEmb

logger = logging.getLogger(__name__)


class GenCondRegressor(torch.nn.Module, ABC):
    r"""Generative Conditional Regressor module.

    This is an abstract module inteded to be used as the backbone of a conditional flow-matching/diffusion process which
    enables sampling from the conditional probability distribution:

    .. math::
        \mathbb{P}(X \mid Z)

    Let :math:`\mathcal{X}=\mathbb{R}^{d_x}`, :math:`\mathcal{Z}=\mathbb{R}^{d_z}`, and
    :math:`\mathcal{Y}=\mathbb{R}^{d_v}`.
    Where :math:`X = [x_0,\ldots,x_{T_x}] \in \mathcal{X}^{T_x}` is the input/data sample composed of a
    trajectory of :math:`T_x` points, and :math:`Z = [z_0,\ldots,z_{T_z}] \in \mathcal{Z}^{T_z}` is the
    conditioning/observation variable composed of :math:`T_z` points.

    The module parameterizes a conditional vector-valued regression map:

    .. math::
        \mathbf{f}_{\mathbf{\theta}}: \mathcal{X}^{T_x} \times \mathcal{Z}^{T_z} \times \mathbb{R}
        \to \mathcal{Y}^{T_x},

    with

    .. math::
        V_k = \mathbf{f}_{\mathbf{\theta}}(X_k, Z, k).

    Where :math:`k` denotes the inference-time optimization timestep (i.e., the step of the flow-matching/diffusion)
    process, :math:`X_k` is the noisy version of the data sample at step `k`, and
    :math:`V_k \in (\mathbb{R}^{d_v})^{T_x}` is the target regression vector-valued variable.
    For diffusion models :math:`V_k` typically corresponds to the score functional of
    :math:`\mathbb{P}_k(X \mid Z)`, while for flow-matching models it typically corresponds
    to the flow-matching velocity vector field.

    This abstract base class does not impose equivariance/invariance constraints by itself.
    """

    def __init__(self, in_dim: int, out_dim: int, cond_dim: int):
        super().__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.cond_dim = cond_dim

    @abstractmethod
    def forward(self, X: torch.Tensor, opt_step: torch.Tensor | float | int, Z: torch.Tensor):
        r"""Forward pass of the generative conditional regressor.

        Args:
            X (:class:`torch.Tensor`): The input/data sample composed of a trajectory of `T_x` points in a
                `d_x`-dimensional space. Shape: `(B, T_x, d_x)`, where `B` is the batch size.
            opt_step (:class:`torch.Tensor` | :class:`float` | :class:`int`): The optimization step(s) `k` at which to
                evaluate the regressor. Can be a single scalar or a tensor of shape `(B,)`.
            Z (:class:`torch.Tensor`): The conditioning/observation variable composed of `T_z` points in a
                `d_z`-dimensional space. Shape: `(B, T_z, d_z)`, where `B` is the batch size.

        Returns:
            :class:`torch.Tensor`: The output regression variable of shape `(B, T_x, d_v)`.
        """
        pass


class CondTransformerRegressor(GenCondRegressor):
    r"""Transformer-based generative conditional regressor.

    The module parameterizes :math:`\mathbf{f}_{\mathbf{\theta}}(X_k, Z, k)` with a stack of Transformer blocks.
    The input trajectory
    :math:`X_k` is first projected into an embedding space and interpreted as the target (`tgt`) sequence of a standard
    :class:`torch.nn.TransformerDecoder`. Conditioning information is packed into the decoder `memory` stream:

    * The inference-time step `k` is mapped with a sinusoidal embedding and inserted as the first conditioning token.
    * The observed sequence :math:`Z` is linearly embedded, receives learned positional encodings, and is appended after
      the step token.
    * When ``n_cond_layers > 0`` the conditioning tokens are processed by a Transformer encoder so that the decoder
      attends to context-aware features; otherwise a lightweight MLP refines the embeddings.

    During decoding, self-attention layers refine :math:`X_k` internally while cross-attention layers pull information
    from the conditioning memory, enabling the model to fuse optimisation step, observations, and trajectory features at
    every layer.

    The model map is:

    .. math::
        \mathbf{f}_{\mathbf{\theta}}:
        \mathbb{R}^{d_x \times T_x} \times \mathbb{R}^{d_z \times T_z} \times \mathbb{R}
        \to \mathbb{R}^{d_v \times T_x}.

    This is an unconstrained baseline model (no explicit group-equivariance constraint).

    Args:
        in_dim (:class:`int`): Dimensionality of each element in :math:`X`.
        out_dim (:class:`int`): Dimensionality of the regressed vector field.
        cond_dim (:class:`int`): Dimensionality of each conditioning element in :math:`Z`.
        in_horizon (:class:`int`): Maximum length of :math:`X`.
        cond_horizon (:class:`int`): Maximum length of :math:`Z` (excluding the optimization-step token).
        num_layers (:class:`int`): Number of Transformer decoder layers.
        num_attention_heads (:class:`int`): Number of attention heads in Multi-Head Attention blocks.
        embedding_dim (:class:`int`): Dimensionality of token embeddings.
        p_drop_emb (:class:`float`): Dropout applied to embeddings.
        p_drop_attn (:class:`float`): Dropout applied inside attention blocks.
        causal_attn (:class:`bool`): Whether to use causal attention in self-attention and cross-attention layers.
        num_cond_layers (:class:`int`): Number of encoder layers dedicated to conditioning tokens.
    """

    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        cond_dim: int,
        in_horizon: int,
        cond_horizon: int,
        num_layers: int = 6,
        num_attention_heads: int = 6,
        embedding_dim: int = 768,
        p_drop_emb: float = 0.1,
        p_drop_attn: float = 0.1,
        causal_attn: bool = False,
        num_cond_layers: int = 0,
    ) -> None:
        super().__init__(in_dim=in_dim, out_dim=out_dim, cond_dim=cond_dim)

        assert cond_horizon > 0, f"{cond_horizon} !> 0"
        assert in_horizon > 0, f"{in_horizon} !> 0"

        self.in_horizon = in_horizon
        self.cond_horizon = cond_horizon + 1  # Inference-time opt step is another token

        # Input embedding stem
        self.input_emb = torch.nn.Linear(in_dim, embedding_dim)
        self.pos_emb = torch.nn.Parameter(torch.zeros(1, self.in_horizon, embedding_dim))
        self.drop = torch.nn.Dropout(p_drop_emb)

        # Conditioning variables z and k embedding stem
        self.cond_emb = torch.nn.Linear(cond_dim, embedding_dim)
        self.cond_pos_emb = torch.nn.Parameter(torch.zeros(1, self.cond_horizon, embedding_dim))
        self.opt_time_emb = SinusoidalPosEmb(embedding_dim)  # Inference-time optimization step embedding

        self.encoder = None
        self.decoder = None

        if num_cond_layers > 0:
            encoder_layer = torch.nn.TransformerEncoderLayer(
                d_model=embedding_dim,
                dim_feedforward=4 * embedding_dim,
                nhead=num_attention_heads,
                dropout=p_drop_attn,
                activation="gelu",
                batch_first=True,
                norm_first=False,
            )
            self.encoder = torch.nn.TransformerEncoder(encoder_layer=encoder_layer, num_layers=num_cond_layers)
        else:
            self.encoder = torch.nn.Sequential(
                torch.nn.Linear(embedding_dim, 4 * embedding_dim),
                torch.nn.Mish(),
                torch.nn.Linear(4 * embedding_dim, embedding_dim),
            )

        # decoder
        decoder_layer = torch.nn.TransformerDecoderLayer(
            d_model=embedding_dim,
            nhead=num_attention_heads,
            dim_feedforward=4 * embedding_dim,
            dropout=p_drop_attn,
            activation="gelu",
            batch_first=True,
            norm_first=False,  # important for stability
        )
        self.decoder = torch.nn.TransformerDecoder(decoder_layer=decoder_layer, num_layers=num_layers)

        # Self-Attention and Cross-Attention mask.
        # Cross-attention is used to compute updates to the action vector based on the conditioing tokens
        # composed of a inference-time optimization step token and the observation conditioning tokens.
        if causal_attn:
            # causal mask to ensure that attention is only applied to the left in the input sequence
            # torch.nn.Transformer uses additive mask as opposed to multiplicative mask in minGPT
            # therefore, the upper triangle should be -inf and others (including diag) should be 0.
            mask = (torch.triu(torch.ones(self.in_horizon, self.in_horizon)) == 1).transpose(0, 1)
            mask = mask.float().masked_fill(mask == 0, float("-inf")).masked_fill(mask == 1, float(0.0))
            self.register_buffer("self_att_mask", mask)

            t, s = torch.meshgrid(torch.arange(self.in_horizon), torch.arange(self.cond_horizon), indexing="ij")
            mask = t >= (s - 1)  # add one dimension since opt-time  is the first token in cond
            mask = mask.float().masked_fill(mask == 0, float("-inf")).masked_fill(mask == 1, float(0.0))
            self.register_buffer("cross_att_mask", mask)
        else:
            self.self_att_mask = None
            self.cross_att_mask = None

        # Decoder head
        self.layer_norm = torch.nn.LayerNorm(embedding_dim)
        self.head = torch.nn.Linear(embedding_dim, out_dim)

        # init
        self.apply(self._init_weights)
        logger.info("number of parameters: %e", sum(p.numel() for p in self.parameters()))

    def _init_weights(self, module):
        ignore_types = (
            torch.nn.Dropout,
            SinusoidalPosEmb,
            torch.nn.TransformerEncoderLayer,
            torch.nn.TransformerDecoderLayer,
            torch.nn.TransformerEncoder,
            torch.nn.TransformerDecoder,
            torch.nn.ModuleList,
            torch.nn.Mish,
            torch.nn.Sequential,
        )
        if isinstance(module, (torch.nn.Linear, torch.nn.Embedding)):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)
            if isinstance(module, torch.nn.Linear) and module.bias is not None:
                torch.nn.init.zeros_(module.bias)
        elif isinstance(module, torch.nn.MultiheadAttention):
            weight_names = ["in_proj_weight", "q_proj_weight", "k_proj_weight", "v_proj_weight"]
            for name in weight_names:
                weight = getattr(module, name)
                if weight is not None:
                    torch.nn.init.normal_(weight, mean=0.0, std=0.02)

            bias_names = ["in_proj_bias", "bias_k", "bias_v"]
            for name in bias_names:
                bias = getattr(module, name)
                if bias is not None:
                    torch.nn.init.zeros_(bias)
        elif isinstance(module, torch.nn.LayerNorm):
            torch.nn.init.zeros_(module.bias)
            torch.nn.init.ones_(module.weight)
        elif isinstance(module, CondTransformerRegressor):
            torch.nn.init.normal_(module.pos_emb, mean=0.0, std=0.02)
            if module.cond_emb is not None:
                torch.nn.init.normal_(module.cond_pos_emb, mean=0.0, std=0.02)
        elif isinstance(module, ignore_types):
            # no param
            pass
        else:
            raise RuntimeError("Unaccounted module {}".format(module))

    def get_optim_groups(self, weight_decay: float = 1e-3):
        """Creates optimizer groups separating out parameters to apply weight decay to and those that don't.

        This long function is unfortunately doing something very simple and is being very defensive:
        We are separating out all parameters of the model into two buckets: those that will experience
        weight decay for regularization and those that won't (biases, and layernorm/embedding weights).
        We are then returning the PyTorch optimizer object.
        """
        # separate out all parameters to those that will and won't experience regularizing weight decay
        decay = set()
        no_decay = set()
        whitelist_weight_modules = (torch.nn.Linear, torch.nn.MultiheadAttention)
        blacklist_weight_modules = (torch.nn.LayerNorm, torch.nn.Embedding)
        for mn, m in self.named_modules():
            for pn, p in m.named_parameters():
                fpn = "%s.%s" % (mn, pn) if mn else pn  # full param name

                if pn.endswith("bias"):
                    # all biases will not be decayed
                    no_decay.add(fpn)
                elif pn.startswith("bias"):
                    # MultiheadAttention bias starts with "bias"
                    no_decay.add(fpn)
                elif pn.endswith("weight") and isinstance(m, whitelist_weight_modules):
                    # weights of whitelist modules will be weight decayed
                    decay.add(fpn)
                elif pn.endswith("weight") and isinstance(m, blacklist_weight_modules):
                    # weights of blacklist modules will NOT be weight decayed
                    no_decay.add(fpn)

        # special case the position embedding parameter in the root GPT module as not decayed
        no_decay.add("pos_emb")
        # no_decay.add("_dummy_variable")
        if self.cond_pos_emb is not None:
            no_decay.add("cond_pos_emb")

        # validate that we considered every parameter
        param_dict = {pn: p for pn, p in self.named_parameters()}
        inter_params = decay & no_decay
        union_params = decay | no_decay
        assert len(inter_params) == 0, "parameters %s made it into both decay/no_decay sets!" % (str(inter_params),)
        assert len(param_dict.keys() - union_params) == 0, (
            "parameters %s were not separated into either decay/no_decay set!"
            % (str(param_dict.keys() - union_params),)
        )

        # create the pytorch optimizer object
        optim_groups = [
            {
                "params": [param_dict[pn] for pn in sorted(list(decay))],
                "weight_decay": weight_decay,
            },
            {
                "params": [param_dict[pn] for pn in sorted(list(no_decay))],
                "weight_decay": 0.0,
            },
        ]
        return optim_groups

    def configure_optimizers(
        self, learning_rate: float = 1e-4, weight_decay: float = 1e-3, betas: tuple[float, float] = (0.9, 0.95)
    ):
        """Creates optimizer groups separating out parameters to apply weight decay to and those that don't."""
        optim_groups = self.get_optim_groups(weight_decay=weight_decay)
        optimizer = torch.optim.AdamW(optim_groups, lr=learning_rate, betas=betas)
        return optimizer

    def forward(self, X: torch.Tensor, opt_step: torch.Tensor | float | int, Z: torch.Tensor):
        r"""Forward pass of the conditional transformer regressor, approximating V_k = f(X_k, Z, k).

        Args:
            X (:class:`torch.Tensor`): The input/data sample composed of a trajectory of `T_x` points in a
                `d_x`-dimensional space. Shape: `(B, T_x, d_x)`, where `B` is the batch size.
            opt_step (:class:`torch.Tensor` | :class:`float` | :class:`int`): The optimization timestep(s) `k` at which
                to evaluate the regressor. Can be a single scalar or a tensor of shape `(B,)`.
            Z (:class:`torch.Tensor`): The conditioning/observation variable composed of `T_z` points in a
                `d_z`-dimensional space. Shape: `(B, T_z, d_z)`, where `B` is the batch size.

        Returns:
            :class:`torch.Tensor`: The output regression variable of shape `(B, T_x, d_v)`.
        """
        # assert X.shape[1] <= self.in_horizon, f"Input horizon {X.shape[1]} larger than {self.in_horizon}"
        # assert Z.shape[1] <= self.cond_horizon - 1, f"Cond horizon {Z.shape[1]} larger than {self.cond_horizon - 1}"

        # 1. Inference-time optimization step embedding (k). First conditioning token.
        if isinstance(opt_step, torch.Tensor):
            opt_steps = opt_step.to(device=X.device, dtype=torch.float32)
        else:
            opt_steps = torch.scalar_tensor(opt_step, device=X.device, dtype=torch.float32)
        # broadcast to batch dimension in a way that's compatible with ONNX/Core ML
        opt_steps = opt_steps.reshape(-1).expand(X.shape[0])
        opt_time_emb = self.opt_time_emb(opt_steps).unsqueeze(1)  # (B,1,n_emb)

        # 2. Conditioning variable Z embedding/tokenization
        z_cond_emb = self.cond_emb(Z)  # (B,Tz,n_emb)
        cond_embeddings = torch.cat([opt_time_emb, z_cond_emb], dim=1)  # (B,Tz + 1,n_emb)
        cond_horizon = cond_embeddings.shape[1]
        cond_pos_emb = self.cond_pos_emb[:, :cond_horizon, :]  # each position maps to a (learnable) vector
        # Transformer encoder of conditing tokens
        cond_tokens = self.drop(cond_embeddings + cond_pos_emb)
        cond_tokens = self.encoder(cond_tokens)  # (B,T_cond,n_emb)

        # 3. Input embedding/tokenization
        input_tokens = self.input_emb(X)

        # 4. Transformer encoder of input tokens with self-attention and cross-attention to cond tokens
        input_horizon = input_tokens.shape[1]
        pos_emb = self.pos_emb[:, :input_horizon, :]  # each position maps to a (learnable) vector
        input_tokens = self.drop(input_tokens + pos_emb)  # (B,Tx,n_emb)
        out_tokens = self.decoder(
            tgt=input_tokens, memory=cond_tokens, tgt_mask=self.self_att_mask, memory_mask=self.cross_att_mask
        )  # (B,Tx,n_emb)

        # 5. Regression head projecting to output dimension.
        out_tokens = self.layer_norm(out_tokens)
        out = self.head(out_tokens)  # (B,Tx, out_dim := d_v)
        return out


def test():  # noqa: D103
    torch.manual_seed(0)
    from tqdm import tqdm

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    print(f"Running test on device: {device}")
    dtype = torch.float32
    torch.set_float32_matmul_precision("high")

    dx, dz, dv = 30, 10, 30
    Tx, Tz = 15, 5
    batch_size = 512
    num_batches = 30

    def build_model():
        model = CondTransformerRegressor(
            in_dim=dx,
            out_dim=dv,
            cond_dim=dz,
            in_horizon=Tx,
            cond_horizon=Tz,
            num_layers=3,
            num_attention_heads=6,
            num_cond_layers=0,
        )
        return model.to(device=device, dtype=dtype).train()

    X_batches = [torch.randn(batch_size, Tx, dx, device=device, dtype=dtype) for _ in range(num_batches)]
    Z_batches = [torch.randn(batch_size, Tz, dz, device=device, dtype=dtype) for _ in range(num_batches)]
    opt_steps = [torch.tensor(float(i % Tx), device=device, dtype=dtype) for i in range(num_batches)]

    def benchmark(model, skip_first: bool = False):
        optimizer = model.configure_optimizers()
        measurements: list[float] = []
        for idx, (x, z, step) in tqdm(enumerate(zip(X_batches, Z_batches, opt_steps))):
            start = torch.cuda.Event(enable_timing=True)
            end = torch.cuda.Event(enable_timing=True)
            optimizer.zero_grad(set_to_none=True)
            start.record()
            out = model(X=x, Z=z, opt_step=step)
            assert out.shape == (batch_size, Tx, dv), f"out shape {out.shape}!= {(batch_size, Tx, dv)}"
            loss = out.mean()
            loss.backward()
            optimizer.step()
            if device.type == "cuda":
                torch.cuda.synchronize()
            end.record()
            torch.cuda.synchronize()
            if skip_first and idx == 0:
                continue
            measurements.append(start.elapsed_time(end) * 1e-3)
        return sum(measurements) / len(measurements)

    eager_model = build_model()
    eager_time = benchmark(eager_model)
    print(f"Eager avg forward+backward step time:    {eager_time:.3f} [s]")

    compiled_model = torch.compile(build_model())
    compiled_time = benchmark(compiled_model, skip_first=True)
    print(f"Compiled avg forward+backward step time: {compiled_time:.3f} [s] (excluding first batch)")


if __name__ == "__main__":
    test()
